import collections
import itertools
from logging import debug
from collections import OrderedDict as odict
import pandas

from CGATReport import Utils


def unique(iterables):
    s = set()
    for x in iterables:
        if x not in s:
            yield x
            s.add(x)


def path2str(path):
    '''convert path to printable string.'''
    if path is None:
        return ""
    if Utils.isString(path):
        return path
    try:
        return "/".join(map(str, path))
    except:
        return str(path)


# This module needs to be properly refactored to use
# proper tree traversal algorithms. It currently is
# a collection of not very efficient hacks.

# The DataTree data structure is discontinued
# - a nested dictionary was more general from a user
# point of view.
class DataTree(object):

    '''a DataTree.

    A data tree is a nested dictionary. A branch or
    leaf is identified by a:term:`path`, a tuple
    of dictionary keys. For example: the:term:path
    ("data1","slice1","distance") returns the
    data at dictionary["data1"]["slice1"]["distance"].

    Note that it will never return a KeyError, but
    will return an empty dictionary for a new tree.
    '''

    slots = "_data"

    def __init__(self, data=None):

        if not data:
            data = odict()
        object.__setattr__(self, "_data", data)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, key):
        return self._data.__getitem__(key)

    def __delitem__(self, key):
        return self._data.__delitem__(key)

    def __setitem__(self, key, value):
        return self._data.__setitem__(key, value)

    def __len__(self):
        return self._data.__len__()

    def getPaths(self):
        '''extract labels from data.

        returns a list of list with all labels within
        the nested dictionary of data.
        '''
        labels = []

        this_level = [self._data, ]

        while 1:
            l, next_level = [], []
            for x in [x for x in this_level if hasattr(x, "keys")]:
                l.extend(list(x.keys()))
                next_level.extend(list(x.values()))
            if not l:
                break
            labels.append(list(unique(l)))
            this_level = next_level

        return labels

    def getLeaf(self, path):
        '''get leaf/branch at *path*.'''
        work = self._data
        for x in path:
            try:
                work = work[x]
            except KeyError:
                work = None
                break
        return work

    def setLeaf(self, path, data):
        '''set leaf/branch at *path* to *data*.'''
        if len(path) == 0:
            object.__setattr__(self, "_data", data)
        else:
            work = self._data
            for x in path[:-1]:
                try:
                    work = work[x]
                except KeyError:
                    work[x] = DataTree()
                    work = work[x]

            work[path[-1]] = data

    def swop(self, level1, level2):
        '''swop two levels *level1* and *level2*.

        For example, swop(0,1) on paths (a/1/x, a/1/y, b/2/x, c/1/y)
        will result in 1/a/x, 1/a/y, 1/c/y, 2/b/x.

        Both levels must be smaller the len().
        '''
        nlevels = len(self.getPaths())
        if nlevels <= level1:
            raise IndexError("level out of range: %i >= %i" %
                             (level1, nlevels))
        if nlevels <= level2:
            raise IndexError("level out of range: %i >= %i" %
                             (level2, nlevels))
        if level1 == level2:
            return
        if level1 > level2:
            level1, level2 = level2, level1

        paths = self.getPaths()
        prefixes = paths[:level1]
        infixes = paths[level1 + 1:level2]
        suffixes = paths[level2 + 1:]

        if prefixes:
            prefixes = list(itertools.product(*prefixes))
        else:
            prefixes = [(None,)]

        if infixes:
            infixes = list(itertools.product(*infixes))
        else:
            infixes = [(None,)]

        if suffixes:
            suffixes = list(itertools.product(*suffixes))
        else:
            suffixes = [(None,)]

        # write to new tree in order to ensure that labels
        # that exist in both level1 and level2 are not
        # overwritten.
        newtree = DataTree()

        def _f(p):
            return tuple([x for x in p if x is not None])

        for p1, p2 in itertools.product(paths[level1], paths[level2]):
            for prefix, infix, suffix in itertools.product(prefixes,
                                                           infixes,
                                                           suffixes):
                oldpath = _f(prefix + (p1,) + infix + (p2,) + suffix)
                newpath = _f(prefix + (p2,) + infix + (p1,) + suffix)
                # note: getLeaf, setLeaf are inefficient in this
                # context as they traverse the tree again
                data = self.getLeaf(oldpath)
                if data is None:
                    continue
                newtree.setLeaf(newpath, data)

        object.__setattr__(self, "_data", newtree._data)

    def removeLeaf(self, path):
        '''remove leaf/branch at *path*.'''
        if len(path) == 0:
            object.__setattr__(self, "_data", odict)
        else:
            work = self._data
            for x in path[:-1]:
                try:
                    work = work[x]
                except KeyError:
                    work = None
                    break
            del work[path[-1]]

    def __str__(self):
        paths = self.getPaths()
        if len(paths) == 0:
            return "NA"
        else:
            return "< datatree: %s >" % str(paths)

    def __getattr__(self, name):
        return getattr(self._data, name)

    def __setattr__(self, name, value):
        setattr(self._data, name, value)


def listAsDataFrame(data, index_title='names',
                    values_are_rows=False):
    '''convert a list of key, value pairs to a dataframe.

    The keys will be a one-level index called *index_title*
    and the values will contain one or more columns.

    If *values_are_rows* is set, keys will be column names
    and values will separate rows in the dataframe.

    '''
    if values_are_rows:
        df = pandas.DataFrame(
            odict(data))
    else:
        df = pandas.DataFrame(
            data=[x[1] for x in data],
            index=pandas.Index([x[0] for x in data],
                               name=index_title))
    return df



def asDataFrame(data):
    '''convert data tree to pandas DataFrame.

    The data frame is multi-indexed according to the depth within the
    data tree.

    If the data-tree has only one level, the data will be
    single-indexed because pandas will not tolerate a single level
    MultiIndex.

    The code assumes that the data tree has a uniform
    depth and structure.

    The inner-most level in the *data* tree will be columns. However,
    if *data* is only a single-level dictionary, the keys in the
    dictionary will be row labels and the resultant dataframe will
    have only one column.

    Depending on the type of the leaf, the data frame is constructed
    as follows:

    Leaves are multiple arrays of the same size

        The assumption is that the data is coordinate type
        data (x,y,z values). Leaves will be added to a dataframe
        as additional columns.

    Leaves are a single array or arrays with dissimilar size

        A melted data frame will be constructed where
        the hierarchical index contains the path information
        and the data frame has a single column with the value.

    Leaf is a dataframe
        Dataframes will be concatenated. Existing indices
        of the dataframes will be preserved with the exception
        of the trivial index for the row numbers.

        Requires:
            All dataframes need to have the same columns.

    Leaf is a scalar
        Dataframes will be built from a nested dictionary

    Special cases for backwards compatibility:

    1. Lowest level dictionary contains the following arrays: rows,
        columns, matrix - numpy matrix, convert to dataframe and apply
        as above

    2. Lowest level dictionary contains the following keys:
        '01', '10', '11' - Venn 2-set data, convert columns
        '001', '010', ... - Venn 3-set data, convert columns

    Pandas attempts to find a column data type that will
    fit all values in a column. Thus, if a column is numeric,
    but contains values such as "inf", "Inf", as well, the
    column type might be set to object or char.

    '''
    if data is None or len(data) == 0:
        return None

    levels = getDepths(data)
    mi, ma = min(levels), max(levels)
    if mi != ma:
        raise NotImplementedError(
            'data tree not of uniform depth, min=%i, max=%i' %
            (mi, ma))

    labels = getPaths(data)

    ######################################################
    ######################################################
    ######################################################
    # check special cases
    MATRIX = ('rows', 'columns', 'matrix')
    VENN2 = ('10', '01', '11')
    VENN3 = ('010', '001', '011')
    dataframe_prune_index = True
    branches = list(getNodes(data, len(labels) - 2))
    for path, branch in branches:
        # numpy matrix - dictionary with keys matrix, rows, columns
        if len(set(branch.keys()).intersection(MATRIX)) == len(MATRIX):
            df = pandas.DataFrame(branch['matrix'],
                                  columns=branch['columns'],
                                  index=branch['rows'])
            setLeaf(data, path, df)
            dataframe_prune_index = False

        elif len(set(branch.keys()).intersection(VENN2)) == len(VENN2) or \
                len(set(branch.keys()).intersection(VENN3)) == len(VENN3):
            # sort so that 'labels' is not the first item
            # specify data such that 'labels' will a single tuple entry
            values = sorted(branch.items())
            df = listAsDataFrame(values)
            dataframe_prune_index = False
            setLeaf(data, path, df)

    ######################################################
    ######################################################
    ######################################################
    labels = getPaths(data)
    # build multi-index
    leaves = list(getNodes(data, len(labels) - 1))

    # if set to a number, any superfluous levels in the
    # hierarchical index of the final dataframe will
    # be removed.
    expected_levels = None

    leaf = leaves[0][1]

    if Utils.isArray(leaf):

        # build dataframe from arrays
        dataframes = []
        index_tuples = []

        # not a nested dictionary
        if len(labels) == 1:
            branches = [(('all',), data)]
        else:
            branches = list(getNodes(data, max(0, len(labels) - 2)))

        # check if it is coordinate data
        # All arrays need to have the same length
        is_coordinate = True
        for path, subtree in branches:
            lengths = [len(x) for x in subtree.values()]
            if len(lengths) == 0:
                continue

            # all arrays have the same length - coordinate data
            if len(lengths) == 1 or min(lengths) != max(lengths):
                is_coordinate = False
                break

        if is_coordinate:
            debug('dataframe conversion: from array - coordinates')
            for path, leaves in branches:
                # skip empty leaves
                if len(leaves) == 0:
                    continue
                dataframes.append(pandas.DataFrame(leaves))
                index_tuples.append(path)
        else:
            debug('dataframe conversion: from array - series')
            # arrays of unequal length are measurements
            # build a melted data frame with a single column
            # given by the name of the path.
            for key, leave in leaves:
                # skip empty leaves
                if len(leave) == 0:
                    continue
                index_tuples.append(key)
                dataframes.append(pandas.DataFrame(leave,
                                                   columns=('value',)))

        expected_levels = len(index_tuples[0])
        # concat will sort the column names if they are not
        # identical between all dataframes. If there are empty
        # dataframes, they count as different and a sort is
        # triggered. Hence make sure there are no empty frames.
        df = pandas.concat(dataframes, keys=index_tuples)

    elif Utils.isDataFrame(leaf):
        debug('dataframe conversion: from dataframe')

        # build dataframe from list of dataframes
        # by concatenation.
        # Existing indices of the dataframes will
        # be added as columns.
        dataframes = []
        index_tuples = []
        path_lengths = []
        levels = []
        for path, dataframe in leaves:
            path_lengths.append(len(path))
            if len(path) == 1:
                # if only one level, do not use tuple
                index_tuples.append(path[0])
            else:
                index_tuples.append(path)
            dataframes.append(dataframe)

            levels.append(Utils.getDataFrameLevels(
                dataframe,
                test_for_trivial=True))

        assert min(path_lengths) == max(path_lengths)
        assert min(levels) == max(levels)

        expected_levels = min(path_lengths) + min(levels)
        df = pandas.concat(dataframes, keys=index_tuples)

        # concat is akin to an SQL join operation and will
        # sort the columns lexicographically.
        # For matrices that could result in the columns not
        # being in the natural order any more unless the
        # column names have been set accordingly.
        # The following thus reorders the columns preserving
        # the order in the original dataframes.
        column_set = collections.OrderedDict()
        for d in dataframes:
            column_set.update(collections.OrderedDict(
                [(x, 0) for x in d.columns]))

        df = df.reindex_axis(column_set.keys(), axis=1)

    else:
        debug('dataframe conversion: from values')
        if len(labels) == 1:
            # { 'x': 1, 'y': 2 } -> DF with one row and two columns (x, y)
            df = pandas.DataFrame(data.values(), index=data.keys())
        elif len(labels) == 2:
            # { 'a': {'x':1, 'y':2}, 'b': {'y',2}
            # -> DF with two columns(x,y) and two rows(a,b)
            df = pandas.DataFrame.from_dict(data).transpose()
            # reorder so that order of columns corresponds to data
            df = df[labels[-1]]
        else:
            # We are dealing with a simple nested dictionary
            branches = list(getNodes(data, max(0, len(labels) - 3)))
            dataframes = []
            index_tuples = []
            for path, nested_dict in branches:
                # transpose to invert columns and rows
                # in cgatreport convention, the deeper
                # level in a dictionary in cgatreport are columns, while
                # in pandas they are rows.
                df = pandas.DataFrame(nested_dict).transpose()
                dataframes.append(df)
                index_tuples.extend([path])
            df = pandas.concat(dataframes, keys=index_tuples)

    # remove index with row numbers
    if expected_levels is not None and dataframe_prune_index:
        Utils.pruneDataFrameIndex(df, expected_levels)

    # rename levels in hierarchical index
    is_hierarchical = isinstance(df.index,
                                 pandas.core.index.MultiIndex)

    if is_hierarchical:
        n = list(df.index.names)
        l = ["track", "slice"] + ["level%i" % x for x in range(len(n))]

        for x, y in enumerate(n):
            if y is None:
                n[x] = l[x]
        df.index.names = n
    else:
        df.index.name = 'track'

    return df


def getPaths(work):
    '''extract labels from data.

    returns a list of list with all labels within
    the nested dictionary of data.
    '''
    labels = []

    this_level = [work, ]

    while 1:
        l, next_level = [], []
        for x in [x for x in this_level if hasattr(x, "keys")]:
            # ignore data frame for path calculation
            if isinstance(x, pandas.DataFrame):
                break
            if isinstance(x, pandas.Series):
                break
            l.extend(list(x.keys()))
            next_level.extend(list(x.values()))
        if not l:
            break
        labels.append(list(unique(l)))
        this_level = next_level

    return labels


def getNodes(work, level=0):
    '''iterate over all nodes at a certain depth.
    in nested dictionary work.

    yields path, value items
    '''

    stack = collections.deque([((), 0, x) for x in work.items()])

    # BFS
    while stack:
        p, l, kv = stack.popleft()
        k, v = kv
        n = p + (k,)
        if l == level:
            yield (n, v)
        if isinstance(v, dict):
            stack.extend([(n, l + 1, x) for x in v.iteritems()])


def getDepths(work):
    '''return a list of depth of leaves.'''
    stack = [(0, x) for x in work.values()]

    levels = []
    while stack:
        level, v = stack.pop()
        if isinstance(v, dict):
            stack.extend([(level + 1, x) for x in v.values()])
        else:
            levels.append(level)
    return levels


def getLeaf(work, path):
    '''get leaf/branch at *path*.'''
    for x in path:
        try:
            work = work[x]
        except KeyError:
            work = None
            break
        except TypeError:
            work = None
            break

    return work


def setLeaf(work, path, data):
    '''set leaf/branch at *path* to *data*.'''
    for x in path[:-1]:
        try:
            work = work[x]
        except KeyError:
            work[x] = odict()
            work = work[x]
    work[path[-1]] = data


def getPrefixes(work, level):
    '''get all possible paths up to *level*.'''
    paths = getPaths(work)
    return list(itertools.product(*paths[:level]))


def removeLevel(work, level):
    '''remove *level* in *work*.'''
    prefixes = getPrefixes(work, level)
    for path in prefixes:
        leaf = getLeaf(work, path)
        # skip truncated branches
        if leaf == None:
            continue

        # delete all in leaf
        keys = list(leaf.keys())
        for key in keys:
            # there might be a subkey the same as key
            d = leaf[key]
            del leaf[key]
            try:
                for subkey, item in d.items():
                    leaf[subkey] = item
            except AttributeError:
                # for items that are not dict
                setLeaf(work, path, d)


def swop(work, level1, level2):
    '''swop two levels *level1* and *level2*.

    For example, swop(0,1) on paths (a/1/x, a/1/y, b/2/y, b/2/x, c/1/y)
    will result in 1/a/x, 1/a/y, 1/c/y, 2/b/a, 2/b/x.

    Both levels must be smaller the len().

    The sort order in lower levels is preserved, i.e.
    it will be 2/b/y, 2/b/x.
    '''
    paths = getPaths(work)
    nlevels = len(paths)
    if nlevels <= level1:
        raise IndexError("level out of range: %i >= %i" % (level1, nlevels))
    if nlevels <= level2:
        raise IndexError("level out of range: %i >= %i" % (level2, nlevels))
    if level1 == level2:
        return
    if level1 > level2:
        level1, level2 = level2, level1

    prefixes = paths[:level1]
    infixes = paths[level1 + 1:level2]
    suffixes = paths[level2 + 1:]

    if prefixes:
        prefixes = list(itertools.product(*prefixes))
    else:
        prefixes = [(None,)]

    if infixes:
        infixes = list(itertools.product(*infixes))
    else:
        infixes = [(None,)]

    if suffixes:
        suffixes = list(itertools.product(*suffixes))
    else:
        suffixes = [(None,)]

    # write to new tree in order to ensure that labels
    # that exist in both level1 and level2 are not
    # overwritten.
    newtree = odict()

    def _f(p):
        return tuple([x for x in p if x is not None])

    for p1, p2 in itertools.product(paths[level1], paths[level2]):

        for prefix, infix in itertools.product(prefixes, infixes):

            w = getLeaf(work, _f(prefix + (p1,) + infix + (p2,)))
            subpaths = getPaths(w)
            if subpaths:
                suffixes = list(itertools.product(subpaths[0]))
            else:
                suffixes = [(None,)]

            for suffix in suffixes:
                oldpath = _f(prefix + (p1,) + infix + (p2,) + suffix)
                newpath = _f(prefix + (p2,) + infix + (p1,) + suffix)

                # note: getLeaf, setLeaf are inefficient in this
                # context as they traverse the tree again
                data = getLeaf(work, oldpath)
                if data is None:
                    continue
                setLeaf(newtree, newpath, data)

    return newtree


def removeLeaf(work, path):
    '''remove leaf/branch at *path*.

    raises KeyError if path is not found.
    '''

    if len(path) == 0:
        work.clear()
    else:
        for x in path[:-1]:
            work = work[x]
        del work[path[-1]]
    return work


def removeEmptyLeaves(work):
    '''traverse data tree in DFS order and remove empty
    leaves.
    '''

    to_delete = []
    try:
        for label, w in work.items():
            keep = removeEmptyLeaves(w)
            if not keep:
                to_delete.append(label)

        for label in to_delete:
            del work[label]

        if len(work) == 0:
            return False
        else:
            return True
    except AttributeError:
        pass

    # return True if not empty
    # numpy arrays do not test True if they contain
    # elements.
    if isinstance(work, pandas.DataFrame) or isinstance(work, pandas.Series):
        return len(work) > 0

    try:
        return work is not None or work != ""
    except ValueError:
        # for numpy arrays
        return len(work) > 0


def prettyprint(work):
    paths = work.getPaths()
    if len(paths) == 0:
        return "NA"
    else:
        return "< datatree: %s >" % str(paths)


def flatten(l, ltypes=(list, tuple)):
    '''flatten a nested list/tuple.'''

    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            if not l[i]:
                l.pop(i)
                i -= 1
                break
            else:
                l[i:i + 1] = l[i]
        i += 1
    return ltype(l)


def count_levels(labels):
    '''count number of levels for each level in labels'''
    counts = []
    for x in labels:
        if type(x[0]) in Utils.ContainerTypes:
            counts.append(len(x[0]))
        else:
            counts.append(1)
    return counts


def tree2table(data, transpose=False, head=None):
    """build table from data.

    The table will be multi-level (main-rows and sub-rows), if:

       1. there is more than one column
       2. each cell within a row is a list or tuple

    If any of the paths contain tuples/lists, these are
    expanded to extra columns as well.

    If head is given, only first head rows are output.

    returns matrix, row_headers, col_headers
    """

    labels = getPaths(data)

    if len(labels) < 2:
        raise ValueError("expected at least two levels for building table, got %i: %s" %
                         (len(labels), str(labels)))

    effective_labels = count_levels(labels)
    # subtract last level (will be expanded) and 1 for row header
    effective_cols = sum(effective_labels[:-1]) - 1

    col_headers = [""] * effective_cols + labels[-1]
    ncols = len(col_headers)

    paths = list(itertools.product(*labels[1:-1]))
    header_offset = effective_cols
    matrix = []

    debug("Datatree.buildTable: creating table with %i columns" %
          (len(col_headers)))

    # the following can be made more efficient
    # by better use of indices
    row_offset = 0
    row_headers = []

    # iterate over main rows
    for x, row in enumerate(labels[0]):

        first = True
        for xx, path in enumerate(paths):

            # get data - skip if there is None
            work = getLeaf(data, (row,) + path)
            if isinstance(work, pandas.DataFrame):
                if work.empty:
                    continue
            else:
                if not work:
                    continue

            row_data = [""] * ncols

            # add row header only for first row (if there are sub-rows)
            if first:
                if type(row) in Utils.ContainerTypes:
                    row_headers.append(row[0])
                    for z, p in enumerate(row[1:]):
                        row_data[z] = p
                else:
                    row_headers.append(row)
                first = False
            else:
                row_headers.append("")

            # enter data for the first row
            for z, p in enumerate(path):
                row_data[z] = p

            # check for multi-level rows
            is_container = True
            max_rows = None
            for y, column in enumerate(labels[-1]):
                if column not in work:
                    continue
                if type(work[column]) not in Utils.ContainerTypes:
                    is_container = False
                    break
                if max_rows == None:
                    max_rows = len(work[column])
                elif max_rows != len(work[column]):
                    raise ValueError("multi-level rows - unequal lengths: %i != %i" %
                                     (max_rows, len(work[column])))

            # add sub-rows
            if is_container:
                # multi-level rows
                for z in range(max_rows):
                    for y, column in enumerate(labels[-1]):
                        try:
                            row_data[
                                y + header_offset] = Utils.quote_rst(work[column][z])
                        except KeyError:
                            pass

                    if z < max_rows - 1:
                        matrix.append(row_data)
                        row_headers.append("")
                        row_data = [""] * ncols
            else:
                # single level row
                for y, column in enumerate(labels[-1]):
                    try:
                        row_data[
                            y + header_offset] = Utils.quote_rst(work[column])
                    except KeyError:
                        pass

            matrix.append(row_data)

            if head and len(matrix) >= head:
                break

    if transpose:
        row_headers, col_headers = col_headers, row_headers
        matrix = list(zip(*matrix))

    # convert headers to string (might be None)
    row_headers = [str(x) for x in row_headers]
    col_headers = [str(x) for x in col_headers]

    return matrix, row_headers, col_headers


def fromCache(cache,
              tracks=None,
              slices=None,
              groupby="slice"):
    '''return a data tree from cache'''

    data = DataTree()
    keys = [x.split("/") for x in list(cache.keys())]

    if tracks == None:
        tracks = set([x[0] for x in keys])
    else:
        tracks = tracks.split(",")

    if slices == None:
        slices = set([x[1] for x in keys if len(x) > 1])
    else:
        slices = slices.split(",")

    def tokey(track, slice):
        return "/".join((track, slice))

    if not slices:
        for track in tracks:
            data[track] = cache[track]
    elif groupby == "slice" or groupby == "all":
        for slice in slices:
            data[slice] = odict()
            for track in tracks:
                data[slice][track] = cache[tokey(track, slice)]
    elif groupby == "track":
        for track in tracks:
            data[track] = odict()
            for slice in slices:
                data[track][slice] = cache[tokey(track, slice)]
    return data


def prune(data,
          ignore=[],
          method='bottom-up'):
    '''prune data tree.

    Remove all empty leaves.

    Remove all levels from the data tree that are
    superfluous, i.e. levels that contain only a single label
    and all labels in the hierarchy below are the same.

    Do not prune top-level, if it is the only level.

    Ignore certain labels given by ignore.

    Returns the levels and labels which have been pruned.
    '''

    # remove all empty leaves
    removeEmptyLeaves(data)

    # prune superfluous levels
    data_paths = getPaths(data)
    nlevels = len(data_paths)

    levels_to_prune = []

    # used to be: 1 (no pruning on first level)
    for level in range(0, nlevels):

        # check for single label in level
        if len(data_paths[level]) == 1:
            label = data_paths[level][0]
            if label in ignore:
                continue
            prefixes = getPrefixes(data, level)
            keep = False
            for prefix in prefixes:
                leaves = getLeaf(data, prefix)
                if leaves is None:
                    continue
                if len(leaves) > 1 or label not in leaves:
                    keep = True
                    break
            if not keep:
                levels_to_prune.append((level, label))

    levels_to_prune.reverse()

    pruned = []

    for level, label in levels_to_prune:
        # do not prune top-level, if it is the only level
        if level == 0 and nlevels == 1:
            continue
        if method == 'bottom-up':
            # stop pruning if not on deepest level
            if level != nlevels - 1:
                break

        pruned.append((level, label))
        removeLevel(data, level)
        nlevels -= 1

    return pruned



import sys
import os
import re
import random
import glob

from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict
from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport import Utils


import matplotlib
from matplotlib import pyplot as plt

from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.numpy2ri


class MatplotlibData(Tracker):

    '''create plot using matplotlib.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)

        # do the plotting
        fig = plt.figure()
        plt.plot(s)
        return odict((("text", "#$mpl %i$#" % fig.number),))


def getCurrentRDevice():
    '''return the numerical device id of the current device.'''
    return R["dev.cur"]()[0]


class RPlotData(Tracker):

    '''create plot using R.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        # do the plotting
        R.x11()
        R.plot(s, s)
        return odict((("text", "#$rpl %i$#" % getCurrentRDevice()),))

IMAGEDIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", "images")


class Images(Tracker):
    tracks = glob.glob(os.path.join(IMAGEDIR, "*.png"))

    def __call__(self, track, slice=None):
        rst_text = '''
This is a preface

.. figure:: %s

Some more text for the figure\n''' % track

        return odict((("rst", rst_text),))


class Images2(Tracker):

    tracks = "all"

    def __call__(self, track, slice=None):

        blocks = ResultBlocks()

        block = '''
.. figure:: %(image)s
:height: 300
'''
        for image in glob.glob(os.path.join(IMAGEDIR, "*.png")):
            blocks.append(ResultBlock(text=block % locals(),
                                      title="image"))

        return odict((("rst", "\n".join(Utils.layoutBlocks(blocks, layout="columns-2"))),))



'''Plotting using the ggplot module.'''

from CGATReportPlugins.Renderer import Renderer
from CGATReportPlugins.Plotter import Plotter
from docutils.parsers.rst import directives

import matplotlib.pyplot as plt
import pandas

# import all into namespace for eval
from ggplot import *


class GGPlot(Renderer, Plotter):

    """Use the python ggplot libary for plotting.
    """
    options = (
        ('aes',  directives.unchanged),
        ('geom', directives.unchanged),
    ) + Renderer.options + Plotter.options

    group_level = 0

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        if "aes" not in kwargs:
            raise ValueError("ggplot renderer requires an aesthetics option")

        if "geom" not in kwargs:
            raise ValueError("ggplot renderer requires a geom option")

        self.aes = kwargs.get('aes')
        self.geom = kwargs.get('geom')

        # at most 6 lines per plot - limit of colours
        # in ggplot
        if self.split_at == 0:
            self.split_at = 6
        else:
            self.split_at = min(self.split_at, 6)

    def render(self, dataframe, path):

        # the index in the dataframe is reset in order
        # to add the index as a column.

        # Currently, ggplot is not using hierarchical indices, but
        # see this thread: https://github.com/yhat/ggplot/issues/285

        dataframe.reset_index(inplace=True)
        s = "p = ggplot(aes(%s), data=dataframe) + %s" % (self.aes, self.geom)

        try:
            exec s in globals(), locals()
        except Exception, msg:
            raise Exception(
                "ggplot raised error for statement '%s': msg=%s" %
                (s, msg))

        # p.draw() calls figure() command, so do not call self.startPlot()
        self.mFigure += 1
        if self.title:
            plt.title(self.title)
        plts = [p.draw()]
        return self.endPlot(plts, None, path)




import numpy as np
import matplotlib.pyplot as plt
from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from collections import OrderedDict as odict

def ExampleWithoutData():
    '''example taken from matplotlib gallery.'''
    N = 5
    menMeans = (20, 35, 30, 35, 27)
    womenMeans = (25, 32, 34, 20, 25)
    menStd = (2, 3, 4, 1, 2)
    womenStd = (3, 5, 2, 3, 3)
    ind = np.arange(N)    # the x locations for the groups
    width = 0.35       # the width of the bars: can also be len(x) sequence

    p1 = plt.bar(ind, menMeans,   width, color='r', yerr=womenStd)
    p2 = plt.bar(ind, womenMeans, width, color='y',
                 bottom=menMeans, yerr=menStd)

    plt.ylabel('Scores')
    plt.title('Scores by group and gender')
    plt.xticks(ind + width / 2., ('G1', 'G2', 'G3', 'G4', 'G5'))
    plt.yticks(np.arange(0, 81, 10))
    plt.legend((p1[0], p2[0]), ('Men', 'Women'))

    # return a place holder for this figure
    return ResultBlocks(
        ResultBlock("#$mpl 1$#\n", ""),
        title="MyTitle")


def ExampleData1():
    return odict((
        ("menMeans", (20, 35, 30, 35, 27),),
        ("womenMeans", (25, 32, 34, 20, 25),),
        ("menStd", (2, 3, 4, 1, 2),),
        ("womenStd", (3, 5, 2, 3, 3)),))


def ExampleData2():
    return odict((
        ("menMeans", (30, 25, 20, 25, 17),),
        ("womenMeans", (15, 12, 24, 10, 15),),
        ("menStd", (2, 1, 2, 1, 2),),
        ("womenStd", (1, 2, 2, 4, 3)),))


def ExampleWithData(data, path):
    '''modified example from matplotlib gallery.'''

    # skip the uppermost levels as we have no
    # tracks or slices
    N = 5
    ind = np.arange(N)    # the x locations for the groups
    width = 0.35       # the width of the bars: can also be len(x) sequence

    p1 = plt.bar(ind,
                 data["menMeans"],
                 width,
                 color='r',
                 yerr=data["womenStd"])
    p2 = plt.bar(ind,
                 data["womenMeans"],
                 width,
                 color='y',
                 bottom=data["menMeans"],
                 yerr=data["menStd"])

    plt.ylabel('Scores')
    plt.title('Scores by group and gender')
    plt.xticks(ind + width / 2., ('G1', 'G2', 'G3', 'G4', 'G5'))
    plt.yticks(np.arange(0, 81, 10))
    plt.legend((p1[0], p2[0]), ('Men', 'Women'))

    # return a place holder for this figure
    return ResultBlocks(
        ResultBlock("#$mpl 0$#\n", ""),
        title="MyTitle")



import os
import sys
import re
import warnings
import inspect
import logging
import glob
import gzip

from collections import OrderedDict as odict
import collections

# Python 2/3 Compatibility
try:
    import ConfigParser as configparser
except:
    import configparser

import numpy
import pandas
import sqlalchemy
import sqlalchemy.exc as exc
import sqlalchemy.engine

# for rpy2 for data frames
try:
    from rpy2.robjects import r as R
except ImportError:
    R = None

from CGATReport import Utils
from CGATReport import Stats


class SQLError(Exception):
    pass


def prettyFloat(val, format="%5.2f"):
    """output a float or "na" if not defined"""
    try:
        x = format % val
    except (ValueError, TypeError):
        x = "na"
    return x


def prettyPercent(numerator, denominator, format="%5.2f"):
    """output a percent value or "na" if not defined"""
    try:
        x = format % (100.0 * numerator / denominator)
    except (ValueError, ZeroDivisionError):
        x = "na"
    return x


def getCallerLocals(level=3, decorators=0):
    '''returns locals of caller using frame.

    optional pass number of decorators

    from http://pylab.blogspot.com/2009/02/python-accessing-caller-locals-from.html
    '''
    f = sys._getframe(level + decorators)
    args = inspect.getargvalues(f)
    return args[3]


def quoteField(s):
    '''returns a quoted version of s for inclusion in SQL statements.'''
    # replace internal "'" with "\'"
    return re.sub("'", "''", s)


@Utils.memoized
def getTableNames(db, database=None, attach=None):
    '''return a set of table names.'''

    # attached databases do not show up in sqlalchemy, thus provide
    # a hack (for sqlite3 only)
    def _get(name):
        try:
            r = db.execute(
                "SELECT tbl_name FROM %s.sqlite_master" % name)
        except exc.SQLAlchemyError as msg:
            raise SQLError(msg)

        return ['%s.%s' % (name, x[0]) for x in r.fetchall()]

    if database is not None:
        return set(_get(database))

    result = []
    # get attached databases
    if attach:
        for path, name in attach:
            result.extend(_get(name))

    inspector = sqlalchemy.engine.reflection.Inspector.from_engine(db)
    result.extend(inspector.get_table_names())
    return set(result)


def getTableColumns(db, tablename, attach=None):
    '''return column information for table *tablename*.

    The returned information contains one dictionary per
    column. The column name is in the "name" field.
    
    If attach is not None, assume that engine is sqlite and
    use a direct query for table names.
    '''
    # for sqlite attached tables
    if attach and "." in tablename:
        # PRAGMA table_info(tablename) does not work as 
        # a "." in a table name will cause an error.
        # Thus select from sqlite_master in attached
        # database directly and parse the result.
        try:
            database, name = tablename.split(".")
            r = db.execute(
                """SELECT sql FROM %s.sqlite_master
                WHERE tbl_name = '%s' AND type = 'table'""" %
                (database, name))
        except exc.SQLAlchemyError as msg:
            raise SQLError(msg)
        # extract string from brackes onwards, ignore CREATE .... (
        s = r.fetchone()[0]
        s = s[s.index("(")+1:]
        # convert to dict for compatibility with sqlalchemy inspector
        vals = [{"name": x} for x in re.findall("(\S+)[^,]+,", s)]
    else:
        inspector = sqlalchemy.engine.reflection.Inspector.from_engine(db)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            vals = inspector.get_columns(tablename)

    return vals


class Tracker(object):

    """
    Base class for trackers. User trackers should be derived from this class.

    A tracker provides the data for a given:term:`track` through
    its __call__ method. Optionally, the data can be sliced, with
    a:term:`slice` containing a subset of the data.

    For example, tracks could be
    entities like cars, motorcycles, and bikes, and slices could
    be colours like blue, red, green::

       class LengthTracker(Tracker):
          mData = ...
          def __call__(self, track, slice):
             return [x.length for x in mData
                      if x.type == track and x.color == slice]

    The call::

       tracker = LengthTracker()
       tracker("car", "blue")

    would return the lengths of blue cars.
    """

    mMinData = 1

    # set to False, if results of tracker should be cached
    cache = True

    # default: empty tracks/slices
    # tracks = []
    # slices = []
    # paths = []

    def __init__(self, *args, **kwargs):
        pass

    # def getTracks(self):
    #     """return a list of all tracks that this tracker provides."""
    #     return self.tracks

    # def getSlices(self):
    #     """return a list of all slices that this tracker provides.
    #     """
    #     return self.slices

    # def getPaths(self):
    #     """return all paths this tracker provides.
    #     """
    #     return self.paths

    def getShortCaption(self):
        """return one line caption.

        The default is to return the first non-blank line of the
        __doc__ string.

        """
        try:
            for line in self.__doc__.split("\n"):
                if line.strip():
                    return line.strip()
        except AttributeError:
            return ""
        return ""

    def __call__(self, track, slice=None):
        """return a data structure for track:param: track and slice:slice:"""
        raise NotImplementedError(
            "Tracker not fully implemented -> __call__ missing")

    def members(self, locals=None):
        '''function similar to locals() but returning member variables of this
        tracker.

        Convenience function for string substitution. If *locals* is
        given (and a dictionary), the dictionary is added to the
        returned dictionary. Entries in *local* take precedence before
        member variables.

        Typical usage is::

           print "my string with %(vars)s" % (self.members(locals())).

        returns a dictionary

        '''
        # skip tracks and slices to avoid recursion
        # todo: do this for the general case
        # 1. subtract property attributes, or
        # 2. subtract members of Tracker()
        l = dict([(attr, getattr(self, attr)) for attr in dir(self)
                  if not isinstance(attr, collections.Callable) and
                  not attr.startswith("__") and attr != "tracks"
                  and attr != "slices"])

        if locals:
            return dict(l, **locals)
        else:
            return l

###########################################################################
###########################################################################
###########################################################################


class TrackerSingleFile(Tracker):

    '''base class for tracker obtaining data from a single file.

    Tracks and slices are defined by the file contents.
    '''

    def __init__(self, *args, **kwargs):
        Tracker.__init__(self, *args, **kwargs)
        if "filename" not in kwargs:
            raise ValueError(
                "TrackerSingleFile requires a:filename: parameter")

        self.filename = kwargs['filename'].strip()


class TrackerMultipleFiles(Tracker):

    '''base class for trackers obtaining data from a multiple files.

    Tracks are names derived from filenames via a
    regular expression.

    This tracker accepts the following parameters:

    :glob:

        glob expression describing the files to include.

    :regex:

        regular expression for extracting a track label from
        a filename. If not given, the complete filename
        path is used.

    '''
    # do not cache as retrieved directly from file
    # and is usually parameterized
    cache = False

    def getTracks(self, subset=None):
        files = sorted(glob.glob(self.glob))
        self.mapTrack2File = {}
        for f in files:
            try:
                track = self.regex.search(f).groups()[0]
            except AttributeError:
                raise ValueError(
                    "filename %s does not match regular expression" % f)

            self.mapTrack2File[track] = f
        return self.mapTrack2File.keys()

    def __init__(self, *args, **kwargs):
        Tracker.__init__(self, *args, **kwargs)
        if "glob" not in kwargs:
            raise ValueError("TrackerMultipleFiles requires a:glob: parameter")

        self.glob = kwargs['glob'].strip()
        self.regex = kwargs.get('regex', '(.*)')

        if '(' not in self.regex:
            raise ValueError(
                "regular expression requires exactly one group enclosed in ()")
        self.regex = re.compile(self.regex)

    def openFile(self, track):
        '''open a file.'''
        filename = self.mapTrack2File[track]
        if filename.endswith(".gz"):
            infile = gzip.open(filename, "r")
        else:
            infile = open(filename, "r")
        return infile


class TrackerTSV(TrackerSingleFile):

    """Base class for trackers that fetch data from an CSV file.

    Each track is a column in the file.
    """

    def __init__(self, *args, **kwargs):
        TrackerSingleFile.__init__(self, *args, **kwargs)

        self._data = None
        self._tracks = None

    def getTracks(self, subset=None):
        if self.filename.endswith(".gz"):
            inf = gzip.open(self.filename, "r")
        else:
            inf = open(self.filename, "r")

        for line in inf:
            if line.startswith("#"):
                continue
            tracks = line[:-1].split("\t")
            break
        inf.close()
        self._tracks = tracks
        return tracks

    def readData(self):
        if self._data is None:
            if self.filename.endswith(".gz"):
                inf = gzip.open(self.filename, "r")
            else:
                inf = open(self.filename, "r")

            data = [x.split()
                    for x in inf.readlines() if not x.startswith("#")]
            inf.close()

            self.data = dict(list(zip(data[0], list(zip(*data[1:])))))

    def __call__(self, track, **kwargs):
        """return a data structure for track:param: track"""
        self.readData()
        return self.data[track]


class TrackerMatrices(TrackerMultipleFiles):
    """Return matrix data from multiple files.
    """

    def __call__(self, track, **kwargs):
        """return a data structure for track:param: track"""

        infile = self.openFile(track)

        dtype = numpy.float

        lines = [l for l in infile.readlines() if not l.startswith("#")]
        infile.close()
        nrows = len(lines) - 1
        col_headers = lines[0][:-1].split("\t")[1:]
        ncols = len(col_headers)
        matrix = numpy.zeros((nrows, ncols), dtype=dtype)
        row_headers = []

        for row, l in enumerate(lines[1:]):
            data = l.split("\t")
            row_headers.append(data[0])
            matrix[row] = numpy.array(data[1:],
                                      dtype=dtype)

        # convert to floats/ints - in dataframe construction
        # columns get sorted lexicographical order if they
        # are strings.
        def _convert(l):
            try:
                return map(int, l)
            except ValueError:
                pass
            try:
                return map(float, l)
            except ValueError:
                pass
            return l

        row_headers = _convert(row_headers)
        col_headers = _convert(col_headers)

        return odict((('matrix', matrix),
                      ('rows', row_headers),
                      ('columns', col_headers)))


class TrackerDataframes(TrackerMultipleFiles):

    '''return dataframe from files.

    By default, the dataframe has no row names.
    If self.index_column is set, the specified column
    will be used as row names.
    '''

    def __init__(self, *args, **kwargs):
        TrackerMultipleFiles.__init__(self, *args, **kwargs)
        self.index_column = kwargs.get('index_column', None)

    def __call__(self, track, **kwargs):

        df = pandas.read_csv(self.openFile(track),
                             sep='\t',
                             header=0,
                             index_col=self.index_column)
        return df


class TrackerImages(Tracker):

    '''Collect image files and arrange them in a gallery.
    '''
    # do not cache as retrieved directly from file
    # and is usually parameterized
    cache = False

    def __init__(self, *args, **kwargs):
        Tracker.__init__(self, *args, **kwargs)
        if "glob" not in kwargs:
            raise ValueError("TrackerImages requires a:glob: parameter")
        self.glob = kwargs["glob"]

    def getTracks(self, subset=None):
        return glob.glob(self.glob)

    def __call__(self, track, **kwargs):
        """return a data structure for track:param: track and slice:slice:"""
        return odict((('name', track), ('filename', track)))


class TrackerSQL(Tracker):

    """Base class for trackers that fetch data from an SQL database.

    The basic tracker identifies tracks as tables that match a
    certain pattern (:attr:`pattern`)

    The pattern should contain at least one group. If there are multiple
    groups, these will be associated as tracks/slices.

    This tracker connects to the database. Each tracker will establish
    its own connection for efficient multi-processing.

    If:attr:`as_tables` is set, the full table names will be returned.
    The default is to apply:attr:`pattern` and return the result.
    """

    pattern = None
    as_tables = False

    def __init__(self, backend=None, attach=[], *args, **kwargs):
        Tracker.__init__(self, *args, **kwargs)

        # connection within python
        self.db = None

        # connection within R
        self.rdb = None

        # attach to additional tables
        self.attach = attach

        if backend is not None:
            # backend given - use it
            self.backend = backend
        else:
            # not defined previously (by mix-in class) get default
            if not hasattr(self, "backend"):
                self.backend = Utils.PARAMS["report_sql_backend"]

        # patch for mPattern and mAsTables for backwards-compatibility
        if hasattr(self, "mPattern"):
            warnings.warn(
                "mPattern is deprecated, use pattern instead",
                DeprecationWarning)
            self.pattern = "(.*)%s" % self.mPattern
        if hasattr(self, "mAsTables"):
            warnings.warn(
                "mAsTables is deprecated, use as_tables instead",
                DeprecationWarning)
            self.as_tables = self.mAsTables

    def connect(self, creator=None):
        """lazy connection function."""

        if not self.db:

            logging.debug("connecting to %s" % self.backend)

            # attach to additional databases (sqlite)
            if self.attach:

                if creator is not None:
                    raise NotImplementedError(
                        'attach not implemented if creator is set')

                if not self.backend.startswith('sqlite'):
                    raise NotImplementedError(
                        'attach only implemented for sqlite backend')

                def _my_creator():
                    # issuing the ATTACH DATABASE into the sqlalchemy
                    # ORM (self.db.execute(...))
                    # does not work. The database is attached, but tables
                    # are not accessible in later
                    # SELECT statements.
                    import sqlite3
                    conn = sqlite3.connect(
                        re.sub("sqlite:///", "", self.backend))
                    for filename, name in self.attach:
                        conn.execute("ATTACH DATABASE '%s' AS %s" %
                                     (os.path.abspath(filename),
                                      name))
                    return conn
                creator = _my_creator

            # creator can not be None.
            if creator:
                db = sqlalchemy.create_engine(self.backend,
                                              echo=False,
                                              creator=creator)
            else:
                db = sqlalchemy.create_engine(self.backend,
                                              echo=False)

            if not db:
                raise ValueError(
                    "could not connect to database %s" % self.backend)

            db.echo = False

            # ignore unknown type BigInt warnings
            # Note that this step can take a while on large databases
            # with many tables and many columns
            if db and False:
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        self.metadata = sqlalchemy.MetaData(
                            db, reflect=True)
                except AttributeError:
                    self.metadata = sqlalchemy.MetaData(db, reflect=True)

            self.db = db

            logging.debug("connected to %s" % self.backend)

    def rconnect(self, creator=None):
        '''open connection within R to database.'''

        if not self.rdb:

            R.library('RSQLite')

            if creator:
                self.rdb = creator()

            else:
                if self.backend.startswith('sqlite'):
                    self.rdb = R.dbConnect(
                        R.SQLite(), dbname=re.sub("sqlite:///./", "",
                                                  self.backend))
                else:
                    raise NotImplementedError(
                        "can not connect to %s in R" % self.backend)

    def getTables(self, pattern=None, database=None):
        """return a list of tables matching a *pattern*.

        This function does not return table views.

        If *database* is given, tables are returned from
        *database*.

        returns a sorted list of table names.
        """
        self.connect()
        sorted_tables = sorted(getTableNames(
            self.db,
            database=database,
            attach=self.attach))

        if pattern:
            rx = re.compile(pattern)
            return [x for x in sorted_tables if rx.search(x)]
        else:
            return sorted_tables

    def getTableNames(self, pattern=None):
        '''return a list of tablenames matching a *pattern*.
        '''
        return self.getTables(pattern)

    def hasTable(self, tablename):
        """return table with name *tablename*."""
        self.connect()
        return tablename in getTableNames(self.db, attach=self.attach)

    def getColumns(self, tablename):
        '''return a list of columns in table *tablename*.'''

        self.connect()
        columns = getTableColumns(self.db, tablename, attach=self.attach)

        return [re.sub("%s[.]" % tablename, "", x['name']) for x in columns]

    def getTracks(self, *args, **kwargs):
        """return a list of all tracks that this tracker provides.

        Tracks are defined as tables matching the attribute
        :attr:`pattern`.
        """
        if self.pattern:
            rx = re.compile(self.pattern)
            tables = self.getTables(pattern=self.pattern)
            if self.as_tables:
                return sorted([x for x in tables])
            else:
                return sorted([rx.search(x).groups()[0] for x in tables])
        else:
            return ["all"]

    def getPaths(self):
        """return all paths this tracker provides.

        Tracks are defined as tables matching the attribute
        :attr:`pattern`.

        """
        if self.pattern:
            rx = re.compile(self.pattern)
            # let getTracks handle a single group
            if rx.groups < 2:
                return None
            tables = self.getTables(pattern=self.pattern)
            parts = [rx.search(x).groups() for x in tables]
            result = []
            for x in range(rx.groups):
                result.append(sorted(set([part[x] for part in parts])))
            return result
        return None

    def execute(self, stmt):
        self.connect()
        try:
            r = self.db.execute(stmt)
        except exc.SQLAlchemyError as msg:
            raise SQLError(msg)
        return r

    def buildStatement(self, stmt):
        '''fill in placeholders in stmt.'''

        kwargs = self.members(getCallerLocals())
        statement = stmt % dict(list(kwargs.items()))
        return statement

    # --------------------------------------------------
    # Functions returning the results of SQL statements
    # as various python structures (dicts, lists, ...)

    def getFirstRow(self, stmt):
        """return a row of values from SQL statement *stmt* as a list.

        The SQL statement is subjected to variable interpolation.

        This function will return the first row from a SELECT statement
        as a list.

        Returns None if result is empty.
        """
        e = self.execute(self.buildStatement(stmt)).fetchone()
        if e:
            return list(e)
        else:
            return None

    def getRow(self, stmt):
        """return a row of values from an SQL statement as dictionary.

        This function will return the first row from a SELECT
        statement as a dictionary of column-value mappings.

        Returns None if result is empty.
        """
        e = self.execute(self.buildStatement(stmt)).fetchone()
        # assumes that values are sorted in ResultProxy.keys()
        if e:
            return odict([x, e[x]] for x in list(e.keys()))
        else:
            return None

    def getValues(self, stmt):
        """return values from SQL statement *stmt* as a list.

        This function will return the first value in each row
        from an SELECT statement.

        Returns an empty list if there is no result.
        """
        e = self.execute(self.buildStatement(stmt)).fetchall()

        if e:
            return [x[0] for x in e]
        return []

    def getAll(self, stmt):
        """return all rows from SQL statement *stmt* as a dictionary.

        This method is deprecated.

        The dictionary contains key/values pairs where keys are
        the selected columns and the values are the results.

        Example: SELECT column1, column2 FROM table
        Example: { 'column1': [1,2,3], 'column2': [2,4,2] }

        Returns an empty dictionary if there is no result.
        """
        # convert to tuples
        e = self.execute(self.buildStatement(stmt))
        columns = list(e.keys())
        d = e.fetchall()
        return odict(list(zip(columns, list(zip(*d)))))

    def get(self, stmt):
        """return all results from an SQL statement as list of tuples.

        Example: SELECT column1, column2 FROM table
        Result: [(1,2),(2,4),(3,2)]

        Returns an empty list if there is no result.
        """
        return self.execute(self.buildStatement(stmt)).fetchall()

    def getDict(self, stmt):
        """return results from SQL statement *stmt* as a dictionary.

        Example: SELECT column1, column2 FROM table
        Result: { 1: 2, 2: 4, 3: 2}

        The first column is taken as the dictionary key.

        This method is useful to return a data structure
        that can be used for matrix visualization.
        """
        # convert to tuples
        e = self.execute(self.buildStatement(stmt))
        columns = list(e.keys())
        result = odict()
        for row in e:
            result[row[0]] = odict(list(zip(columns[1:], row[1:])))
        return result

    # -------------------------------------
    # Convenience functions for access
    def getValue(self, stmt):
        """returns a single value from SQL statement *stmt*.

        If the result is empty, None is returned.

        The SQL statement is subjected to variable interpolation.

        This function will return the first value in the first row
        from a SELECT statement.
        """
        statement = self.buildStatement(stmt)
        result = self.execute(statement).fetchone()
        if result is None:
            return result
        return result[0]

    def getIter(self, stmt):
        '''returns an iterator over results of SQL statement *stmt*.
        '''
        return self.execute(stmt)

    def getDataFrame(self, stmt):
        '''return results of SQL statement as an pandas dataframe.
        '''
        e = self.execute(self.buildStatement(stmt))
        # the conversion to a list incurs a performance penalty
        # but pandas requires a __len__ method.
        return pandas.DataFrame.from_records(
            list(e),
            columns=e.keys())

    # # -------------------------------------
    # # Direct access functios for return to CGATReport
    # def getRows(self, stmt):
    #     """return all results from an SQL statement as list of tuples.

    #     Example: SELECT column1, column2 FROM table
    #     Result: [(1,2),(2,4),(3,2)]

    #     Returns an empty list if there is no result.
    #     """
    #     return self.execute(self.buildStatement(stmt)).fetchall()

    # def get(self, stmt):
    #     """deprecated - use getRows instead."""
    #     return self.getRows(self.buildStatement(stmt))

    # def getAll(self, stmt):
    #     '''return results of SQL statement as pandas dataframe.
    #     '''
    #     return self.getDataFrame(self.buildStatement(stmt))

    # def getValues(self, stmt):
    #     '''return results of SQL statement as pandas Series.
    #     '''
    #     e = self.exectute(self.buildStatement(stmt))
    #     return pandas.Series([x[0] for x in e])

    # def getRow(self, stmt):
    #     '''return results of SQL statement as pandas dataframe
    #     '''
    #     e = self.execute(self.buildStatement(stmt))
    #     return pandas.Series(e.fetchone())

    # def getFirstRow(self, stmt):
    #     '''return first row of SQL statement as pandas Series.
    #     '''
    #     e = self.execute(self.buildStatement(stmt))
    #     return pandas.Series(e.fetchone())


class TrackerSQLCheckTables(TrackerSQL):
    """Tracker that examines the presence/absence of a certain
    field in a list of tables.

    Define the following attributes:

:attr:`pattern`: pattern to define tracks (see:class:`TrackerSql`)
:attr:`slices` columns to look for in tables

    """

    slices = None
    pattern = None
    as_tables = True

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def __call__(self, track, slice=None):
        """count number of unique occurances of field *slice* in tables
        matching *track*."""

        if slice not in self.getColumns(track):
            return None

        return self.getValue(
            "SELECT COUNT(DISTINCT %(slice)s) FROM %(track)s")


class TrackerSQLCheckTable(TrackerSQL):

    """Tracker that counts existing entries in a table.

    Define the following attributes:
:attr:`exclude_pattern`: columns to exclude
:attr:`pattern`: pattern to define tracks (see:class:`TrackerSql`)
    """

    exclude_pattern = None
    pattern = "(.*)_evol$"

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def __call__(self, track, *args):
        """count number of entries in a table."""

        statement = "SELECT COUNT(%s) FROM %s WHERE %s IS NOT NULL"

        if self.exclude_pattern:
            fskip = lambda x: re.search(self.exclude_pattern, x)
        else:
            fskip = lambda x: False

        tablename = track + "_evol"
        columns = self.getColumns(tablename)
        data = []

        for column in columns:
            if fskip(column):
                continue
            data.append(
                (column, self.getValue(statement % (column,
                                                    tablename,
                                                    column))))
        return odict(data)


class Config(Tracker):

    '''Tracker providing config values of ini files.

    The .ini files need to be located in the directory
    from which cgatreport is called.

    returns a dictionary of key,value pairs.
    '''
    tracks = glob.glob("*.ini")

    def __init__(self, *args, **kwargs):
        Tracker.__init__(self, *args, **kwargs)

    def __call__(self, track, *args):
        """count number of entries in a table."""

        config = configparser.ConfigParser()
        config.readfp(open(track), "r")

        result = odict()

        def convert(value):
            '''convert a value to int, float or str.'''
            rx_int = re.compile("^\s*[+-]*[0-9]+\s*$")
            rx_float = re.compile("^\s*[+-]*[0-9.]+[.+\-eE][+-]*[0-9.]*\s*$")

            if value is None:
                return value

            if rx_int.match(value):
                return int(value), "int"
            elif rx_float.match(value):
                return float(value), "float"
            return value, "string"

        for section in config.sections():
            x = odict()
            for key, value in config.items(section):
                x[key] = odict(list(zip(("value", "type"), convert(value))))
            result[section] = x

        return result


class Empty(Tracker):
    '''Empty tracker

    This tracker servers as placeholder for plots that require no
    input from a tracker.

    '''

    def getTracks(self, subset=None):
        """return a list of all tracks that this tracker provides."""
        return ["empty"]

    def __call__(self, *args):
        return None


class Status(TrackerSQL):

    '''Tracker returning status information.

    Define tracks and slices. Slices will be translated into
    calls to member functions starting with 'test'.

    Each test function should return a tuple with the test
    status and some information.

    If this tracker is paired with a:class:`Renderer.Status`
    renderer, the following values of a test status will be
    translated into icons: ``PASS``, ``FAIL``, ``WARNING``, ``NOT AVAILABLE``.

    The docstring of the test function is used as description.
    '''

    def getSlices(self, subset=None):
        return [x[4:] for x in dir(self) if x.startswith("test")]

    def __call__(self, track, slice):
        if not hasattr(self, "test%s" % slice):
            raise NotImplementedError("test%s not implement" % slice)

        status, value = getattr(self, "test%s" % slice)(track)
        description = getattr(self, "test%s" % slice).__doc__

        return odict(
            (('name', slice),
             ('status', status),
             ('info', str(value)),
             ('description', description)))


class SQLStatementTracker(TrackerSQL):
    '''Tracker representing the result from an SQL statement.
    '''
    statement = None

    # columns to use as tracks
    fields = ("track",)

    def setIndex(self, dataframe):
        '''sets the index according to the fields given.'''
        # removes the "all" index and others
        dataframe.reset_index(drop=True, inplace=True)
        # set index
        try:
            dataframe.set_index(list(self.fields), inplace=True)
        except KeyError, msg:
            raise KeyError(
                "%s: have %s" %
                (msg, dataframe.columns))

    def __call__(self):
        if self.statement is None:
            raise NotImplementedError(
                "statement needs to be set for SQLStatementTracker")
        return self.getAll(self.statement)


class SingleTableTrackerRows(TrackerSQL):

    '''Tracker representing a table with multiple tracks.

    Returns a dictionary of values.

    The tracks are given by rows in table :py:attr:`table`. The tracks are
    specified by the:py:attr:`fields`.

    :py:attr:`fields` is a tuple of column names (default = ``(track,)``).

    If multiple columns are specified, they will all be used to define
    the tracks in the table.

    Rows in the table need to be unique for any
    combination:py:attr:`fields`.

    :py:attr:`extra_columns` can be used to add additional columns to
    the table. This attribute is a dictionary, though only the keys
    are being used.

    :py:attr:`where` is an optional where that will be added as a
    WHERE clause to the SQL statement

    '''
    exclude_columns = ()
    table = None
    fields = ("track",)
    extra_columns = {}
    sort = None
    loaded = False
    where = None

    # not called by default as Mixin class
    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def _load(self):
        '''load data.

        The data is pre-loaded in order to avoid multiple random access
        operations on the same table.
        '''
        if not self.loaded:
            nfields = len(self.fields)

            if self.sort:
                sort_statement = "ORDER BY %s" % self.sort
            else:
                sort_statement = ""

            if self.where:
                where_statement = "WHERE %s" % self.where
            else:
                where_statement = ""
            self._tracks = self.get("SELECT DISTINCT %s FROM %s %s %s" %
                                    (",".join(self.fields),
                                     self.table,
                                     where_statement,
                                     sort_statement))
            columns = self.getColumns(self.table)
            self._slices = [x for x in columns
                            if x not in self.exclude_columns
                            and x not in self.fields] +\
                list(self.extra_columns.keys())
            # remove columns with special characters (:, +, -,)
            self._slices = [x for x in self._slices
                            if not re.search("[:+-]", x)]

            data = self.get("SELECT %s, %s FROM %s" %
                            (",".join(self.fields),
                             ",".join(self._slices),
                             self.table))
            self.data = odict()
            for d in data:
                tr = tuple(d[:nfields])
                self.data[tr] = odict(list(zip(self._slices,
                                               tuple(d[nfields:]))))
        self.loaded = True

    @property
    def tracks(self):
        if not self.hasTable(self.table):
            return []
        if not self.loaded:
            self._load()
        if len(self.fields) == 1:
            return tuple([x[0] for x in self._tracks])
        else:
            return tuple([tuple(x) for x in self._tracks])

    @property
    def slices(self):
        if not self.hasTable(self.table):
            return []
        if not self.loaded:
            self._load()
        return self._slices

    def __call__(self, track, slice=None):
        if not self.loaded:
            self._load()
        if len(self.fields) == 1:
            track = (track,)
        return self.data[track][slice]


class SingleTableTrackerColumns(TrackerSQL):

    '''Tracker representing a table with multiple tracks.

    Returns a dictionary of two sets of data, one given
    by:py:attr:`column` and one for a track.

    The tracks are derived from all columns in
    table:py:attr:`table`. By default, all columns are taken as tracks
    apart from:py:attr:`column` and those listed
    in:py:attr:`exclude_columns`.

    An example for a table using this tracker would be::

       bin   mouse_counts    human_counts
       100   10              10
       200   20              15
       300   10              4

    In the example above, the tracks will be ``mouse_counts`` and
    ``human_counts``. The slices will be ``100``, ``200``, ``300``

    Tracker could be defined as::

       class MyTracker(SingleTableTrackerColumns):
          table = 'mytable'
          column = 'bin'

    '''
    exclude_columns = ("track,")
    table = None
    column = None

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    @property
    def tracks(self):
        if not self.hasTable(self.table):
            return []
        columns = self.getColumns(self.table)
        return [x for x in columns
                if x not in self.exclude_columns and x != self.column]

    @property
    def slices(self):
        if self.column:
            return self.getValues("SELECT DISTINCT %(column)s FROM %(table)s")
        else:
            return []

    def __call__(self, track, slice=None):
        if slice is not None:
            data = self.getValue(
                "SELECT %(track)s FROM %(table)s WHERE %(column)s = '%(slice)s'")
        else:
            data = self.getValues("SELECT %(track)s FROM %(table)s")
        return data


class SingleTableTrackerEdgeList(TrackerSQL):

    '''Tracker returning values from a table with matrix type data
    that is stored in an edge list, for example::

         row   column    value
         A     B         1
         A     C         2
         B     C         3

    Returns a dictionary of values.

    The tracks are given by entries in the:py:attr:`row` column in a
    table:py:attr:`table`.  The slices are given by entries in the
    :py:attr:`column` column in a table.

    The:py:attr:`value` is a third column specifying the value
    returned. If:py:attr:`where` is set, it is added to the SQL
    statement to permit some filtering.

    If:py:attr:`transform` is set, it is applied to the value.

    if:py:attr:`value2` is set, the matrix is assumed to be stored in
    the format ``(row, column, value, value1)``, where ``value`` is
    the value for ``row,col`` and value2 is the value for ``col,row``.

    This method is inefficient, particularly so if there are no
    indices on:py:attr:`row` and:py:attr:`column`.


    '''
    table = None
    row = None
    column = None
    value = None
    value2 = None
    transform = None
    where = "1"

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    @property
    def tracks(self):
        if self.table is None:
            raise NotImplementedError("table not defined")
        if not self.hasTable(self.table):
            raise ValueError("unknown table %s" % self.table)

        if self.value2 is not None:
            return sorted(set(
                self.getValues("SELECT DISTINCT %(row)s FROM %(table)s") +
                self.getValues("SELECT DISTINCT %(column)s FROM %(table)s")))
        else:
            return self.getValues("SELECT DISTINCT %(row)s FROM %(table)s")

    @property
    def slices(self):
        if self.value2 is not None:
            return self.tracks
        else:
            return self.getValues("SELECT DISTINCT %(column)s FROM %(table)s")

    def __call__(self, track, slice=None):

        try:
            val = self.getValue("""SELECT %(value)s FROM %(table)s
            WHERE %(row)s = '%(track)s' AND
            %(column)s = '%(slice)s' AND %(where)s""")
        except exc.SQLAlchemyError:
            val = None

        if val is None and self.value2:
            try:
                val = self.getValue(
                    """SELECT %(value2)s FROM %(table)s
                    WHERE %(row)s = '%(slice)s' AND
                    %(column)s = '%(track)s' AND %(where)s""")
            except exc.SQLAlchemyError:
                val = None

        if val is None:
            return val

        if self.transform:
            return self.transform(val)
        return val


class SingleTableTrackerEdgeListToMatrix(TrackerSQL):

    '''Tracker returning values from a table with matrix type data
    that is stored in an edge list, for example::

         row   column    value
         A     B         1
         A     C         2
         B     C         3

    Returns a numpy matrix.

    The tracks are given by entries in the:py:attr:`row` column in a
    table:py:attr:`table`.  The slices are given by entries in the
    :py:attr:`column` column in a table.

    The:py:attr:`value` is a third column specifying the value
    returned. If:py:attr:`where` is set, it is added to the SQL
    statement to permit some filtering.

    If:py:attr:`transform` is set, it is applied to the value.

    if:py:attr:`value2` is set, the matrix is assumed to be stored in
    the format ``(row, column, value, value1)``, where ``value`` is
    the value for ``row,col`` and value2 is the value for ``col,row``.
    '''
    table = None
    row = None
    column = None
    value = None
    value2 = None
    missing_value = 0
    diagonal_value = 0
    # set to true if matrix is symmetric
    is_symmetric = False
    dtype = numpy.int
    where = "1"

    # saved by preloading 
    _matrix = None
    _rows = None
    _cols = None

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def getTracks(self):
        self._load()
        return self._rows

    def getSlices(self):
        self._load()
        return self._cols

    def _load(self):
        '''load data.

        The data is pre-loaded in order to avoid multiple random access
        operations on the same table.
        '''
        if self._matrix is None:
            if self.value2 is None:
                data = self.get(
                    "SELECT %(row)s, %(column)s, %(value)s "
                    "FROM %(table)s WHERE %(where)s")
            else:
                data = self.get(
                    "SELECT %(row)s, %(column)s, %(value)s, %(value2)s "
                    "FROM %(table)s WHERE %(where)s")

            self._matrix, self._rows, self._cols = \
                Stats.buildMatrixFromEdges(
                    data,
                    is_symmetric=self.is_symmetric,
                    missing_value=self.missing_value,
                    diagonal_value=self.diagonal_value,
                    dtype=self.dtype)
            self._map_rows = dict([(y, x) for x, y in enumerate(self._rows)])
            self._map_cols = dict([(y, x) for x, y in enumerate(self._cols)])

    def __call__(self, track, slice):
        self._load()
        return self._matrix[self._map_rows[track], self._map_cols[slice]]


class MultipleTableTrackerEdgeList(TrackerSQL):

    '''Tracker representing multiple tables with matrix type data.

    Returns a dictionary of values.

    The tracks are given by table names mathing:py:attr:`pattern`.
    '''
    row = None
    column = None
    value = None
    as_tables = True

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def __call__(self, track, slice=None):

        if self.column is None:
            raise ValueError('MultipleTrackerEdgeList requires a column field')
        if self.row is None:
            raise ValueError('MultipleTrackerEdgeList requires a row field')
        if self.value is None:
            raise ValueError('MultipleTrackerEdgeList requires a value field')

        data = self.get("""SELECT %(row)s, %(column)s, %(value)s
                            FROM %(track)s
                            ORDER BY fdr,power""")
        result = odict()
        for row, col, value in data:
            try:
                result[row][col] = value
            except KeyError:
                result[row] = odict()
                result[row][col] = value

        return result


class SingleTableTrackerHistogram(TrackerSQL):

    '''Tracker representing a table with multiple tracks.

    Returns a dictionary of two sets of data, one given
    by:py:attr:`column` and one for a track.

    The tracks are derived from all columns in
    table:py:attr:`table`. By default, all columns are taken as tracks
    apart from:py:attr:`column` and those listed
    in:py:attr:`exclude_columns`.

    An example for a table using this tracker would be::

       bin   mouse_counts    human_counts
       100   10              10
       200   20              15
       300   10              4

    In the example above, the tracks will be ``mouse_counts`` and
    ``human_counts``. The Tracker could be defined as::

       class MyTracker(SingleTableTrackerHistogram):
          table = 'mytable'
          column = 'bin'

    '''
    exclude_columns = ("track,")
    table = None
    column = None
    value = 'data'

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    @property
    def tracks(self):

        if self.column is None:
            raise NotImplementedError(
                "column not set - Tracker not fully implemented")
        if not self.hasTable(self.table):
            return []
        columns = self.getColumns(self.table)
        return [x for x in columns
                if x not in self.exclude_columns and x != self.column]

    def __call__(self, track, slice=None):
        if self.column is None:
            raise NotImplementedError(
                "column not set - Tracker not fully implemented")
        # labels need to be consistent in order
        # so rename track to value.
        data = self.getAll(
            "SELECT %(column)s, %(track)s AS %(value)s FROM %(table)s")
        return data


class MultipleTableTrackerHistogram(TrackerSQL):

    '''Tracker representing multiple table with multiple slices.

    Returns a dictionary of two sets of data, one given
    by:py:attr:`column` and one for a track.

    The tracks are derived from all columns in
    table:py:attr:`table`. By default, all columns are taken as tracks
    apart from:py:attr:`column` and those listed
    in:py:attr:`exclude_columns`.

    An example for a table using this tracker would be::

       bin   mouse_counts    human_counts
       100   10              10
       200   20              15
       300   10              4

    In the example above, the tracks will be ``mouse_counts`` and
    ``human_counts``. The Tracker could be defined as::

       class MyTracker(ManyTableTrackerHistogram):
          pattern = '(.*)_table'
          column = 'bin'

    '''
    exclude_columns = ("track,")
    as_tables = True
    column = None

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    @property
    def slices(self):
        if self.column == None:
            raise NotImplementedError(
                "column not set - Tracker not fully implemented")
        columns = set()
        for table in self.getTableNames(self.pattern):
            columns.update(self.getColumns(table))
        return [x for x in columns if x not in self.exclude_columns and x != self.column]

    def __call__(self, track, slice):
        # check if column exists in particular table - if not, return no data
        if slice not in self.getColumns(track):
            return None

        return self.getAll("""SELECT %(column)s, %(slice)s FROM %(track)s""")

    # def __call__(self, track):

    #     if self.column == None: raise NotImplementedError("column not set - Tracker not fully implemented")
    # get columns in the alphabetical order
    #     columns = sorted(self.getColumns(track) )
    #     if self.column not in columns: raise ValueError("column '%s' missing from '%s'" % (self.column, track))
    #     columns = ",".join([ x for x in columns if x not in self.exclude_columns and x != self.column ])
    # return self.getAll("""SELECT %(column)s, %(columns)s FROM %(track)s""")


class TrackerSQLMulti(TrackerSQL):

    '''An SQL tracker spanning multiple databases.

    '''

    databases = ()
    tracks = ()

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

        if len(self.tracks) == 0:
            raise ValueError("no tracks specified in TrackerSQLMulti")
        if (len(self.tracks) != len(self.databases)):
            raise ValueError(
                "TrackerSQLMulti requires an equal number of "
                "tracks (%i) and databases (%i)"
                % (len(self.tracks), len(self.databases)))
        if not self.backend.startswith("sqlite"):
            raise ValueError("TrackerSQLMulti only works for sqlite database")

        if not self.db:
            def _my_creator():
                # issuing the ATTACH DATABASE into the sqlalchemy ORM
                # (self.db.execute(...))  does not work. The database
                # is attached, but tables are not accessible in later
                # SELECT statements.
                import sqlite3
                conn = sqlite3.connect(re.sub("sqlite:///", "", self.backend))
                for track, name in zip(self.databases, self.tracks):
                    conn.execute("ATTACH DATABASE '%s/csvdb' AS %s" %
                                 (os.path.abspath(track),
                                  name))
                return conn

            self.connect(creator=_my_creator)


class TrackerMultipleLists(TrackerSQL):

    '''A class to retrieve multiple columns across one or more tables.
    Returns a dictionary of lists.

    TrackerMultipleLists can be used in conjunction with venn and
    hypergeometric transformers and the venn render.

    The items in each list are specified by an SQL statement. The
    statements can be specified in 3 different ways:

    :attr:`statements` dictionary
        If the tracker contains a statements attribute then the statments
        are taken from here as well as the list names e.g.::

        class TrackerOverlapTest1(TrackerOverlappingSets):
            statements = {"listA": "SELECT gene_id FROM table_a",
                          "listB": "SELECT gene_id FROM table_b"}

    :attr:`listA`,:attr:`listB`,:attr:`listC` and:attr:`background` attributes

        If the tracker does not contain a statements dictionary then
        the statements can be specifed using these attributes. An
        optional list of labels can be specified for the names of
        these lists. For example::

            class TrackerOverlapTest2(TrackerOverlappingSets):
               listA = "SELECT gene_id FROM table_a"
               listB = "SELECT gene_id FROM table_b"

               labels = ["FirstList","SecondList"]

    :meth:`getStatements` method

         The:meth:`getStatements` method can be overridden to allow
         full control over where the statements come from. It should
         return a dictionary of SQL statements.

    Because TrackerMultipleLists is derived from:class:`TrackerSQL`,
    tracks and slices can be specified in the usual way.

    '''

    statements = None
    ListA = None
    ListB = None
    ListC = None
    background = None
    labels = None

    def getStatements(self):

        statements = odict()

        if self.statements:
            return self.statements

        if self.ListA:
            if self.labels:
                label = self.labels[0]
            else:
                label = "ListA"
            statements[label] = self.ListA

        if self.ListB:

            if self.labels:
                label = self.labels[1]
            else:
                label = "ListB"

            statements[label] = self.ListB

        if self.ListC:

            if self.labels:
                label = self.labels[2]
            else:
                label = "ListC"

            statements[label] = self.ListC

        if self.background:
            statements["background"] = self.background

        return statements

    def __call__(self, track, slice=None):

        statements = self.getStatements()
        # track and slice will be substituted in the statements
        return odict([(
            x, 
            self.getValues(statements[x])) for x in statements])


class MeltedTableTracker(TrackerSQL):

    '''Tracker representing multiple tables with the same columns.

    The tables are melted - a column called ``track`` is added
    that contains the table name (see :py:attr:`column_name`)

    '''
    tracks = "all"
    pattern = None
    column_name = 'track'

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def __call__(self, track):
        assert(self.pattern is not None)
        tables = self.getTables(self.pattern)

        if len(tables) == 0:
            return None

        ref_columns = self.getColumns(tables[0])
        fields = ",".join(ref_columns)
        results = []
        for table in tables:
            columns = self.getColumns(table)
            if columns != ref_columns:
                warnings.warn(
                    "incompatible column names in table %s - skipped" % table)
                continue

            track = re.search(self.pattern, table).groups()[0]

            results.extend(
                self.get(
                    "SELECT '%(track)s' as track, %(fields)s FROM %(table)s"))

        ref_columns.insert(0, self.column_name)

        return odict(zip(ref_columns, zip(*results)))


class MeltedTableTrackerDataframe(MeltedTableTracker):

    '''Tracker representing multiple tables with the same columns.

    The tables are melted - a column called ``track`` is added
    that contains the table name.

    This tracker returns a dataframe directly.
    '''
    tracks = "all"
    pattern = None

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self, *args, **kwargs)

    def __call__(self, track):
        data = MeltedTableTracker.__call__(self, track)
        return pandas.DataFrame.from_dict(data)



'''This document contains a number of sample trackers.'''

import os
import random
import glob
import itertools
import numpy

from CGATReport.Tracker import Tracker, Status
from collections import OrderedDict as odict
import collections


def BarData():
    return dict([("bar1", 20), ("bar2", 10)])


class LabeledDataExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):

        v = int(track[-1]) * int(slice[-1])
        if slice == "slice1":
            return odict((("column1", v),
                          ("column2", v * 2),))
        elif slice == "slice2":
            return odict((("column1", v),
                          ("column2", v * 2),
                          ("column3", v * 3),))


class LabeledDataLargeExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ["track%i" % x for x in range(0, 100)]

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((("column1", 10),
                          ("column2", 20),))
        elif slice == "slice2":
            return odict((("column1", 20),
                          ("column2", 10),
                          ("column3", 5),))



class LabeledDataWithErrorsExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 1),
                                   ))),
                ("column2", odict((('data', 10),
                                   ('error', 2),
                                   ))),
            ))
        elif slice == "slice2":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 3),
                                   ))),
                ("column2", odict((('data', 10),
                                   ('error', 4)))),
                ("column3", odict((('data', 30),
                                   ('error', 5)))),
            ))


class LabeledDataWithErrorsAndLabelsExample(Tracker):

    '''use of ordered dictionary important as first
    element is treated as data.
    '''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 5),
                                   ('label', '**')))),
                ("column2", odict((('data', 10),
                                   ('error', 2),
                                   ('label', '*'))))
            ))
        elif slice == "slice2":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 5),
                                   ('label', '***')))),
                ("column2", odict((('data', 10),
                                   ('error', 1)))),
                ("column3", odict((('data', 30),
                                   ('error', 4)))),
            ))


class HierarchicalLabeledDataExample(Tracker):
    '''labelled data with more than three levels.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):

        nvalues = 4
        values = collections.defaultdict(dict)
        levels = 3
        for x in range(nvalues):
            values[x] = collections.defaultdict(dict)
            for y in range(nvalues):
                values[x][y] = collections.defaultdict(dict)

        if slice == "slice1":
            for a, b, c in itertools.permutations(
                    range(nvalues),
                    levels):
                values[a][b][c] = (a + 1) * (b + 1) * (c + 1)
        elif slice == "slice2":
            for a, b, c in itertools.permutations(
                    range(nvalues),
                    levels):
                values[a][b][c] = (a + 1) * (b + 1) * (c + 1) * 2

        return values


class SingleColumnDataExample(Tracker):

    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class ArrayDataExample(Tracker):

    '''return two arrays of data.'''

    slices = ["slice%i" % x for x in range(0, 2)]
    tracks = ["track%i" % x for x in range(0, 3)]

    def __call__(self, track, slice=None):

        scale = (3 - int(track[-1])) * (2 - int(slice[-1]))

        data = odict((("x", range(0, 10)),
                      ("y", [x * scale for x in range(0, 10)])))

        return data


class SingleColumnDataLargeExample(Tracker):

    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ["track%i" % x for x in range(0, 20)]

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class SingleColumnDataWithErrorExample(Tracker):

    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        e = [random.randint(0, 3) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),
                      ("error", e)))


class SingleColumnDataExampleWithoutSlices(Tracker):

    '''return a single column of data.'''
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class MultipleColumnDataExample(Tracker):

    '''multiple columns each with a column with data.'''
    mColumns = ["col1", "col2", "col3"]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns) - 1):
                data.append([y + random.gauss(0, 0.2) for y in range(10)])
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.5) for y in range(10)])
        return odict(zip(self.mColumns, data))


class MultipleColumnDataFullExample(Tracker):

    '''multiple columns each with a column with data.'''
    mColumns = ["col1", "col2", "col3"]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.2) for y in range(20)])
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.5) for y in range(20)])
        return odict(zip(self.mColumns, data))


class ErrorInTracker1(Tracker):

    '''A tracker that creates an error - exception while collecting data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        raise ValueError("testing: could not collect data")


class ErrorInTracker2(Tracker):

    '''A tracker that creates an error - problems while returning tracks.'''
    slices = ("slice1", "slice2")

    def getTracks(self):
        raise ValueError("testing: could not build tracks")

    def __call__(self, track, slice=None):
        return odict((("data", range(0, 10)),))


class ErrorInTracker3(Tracker):

    '''A tracker that creates an error - no data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        return None


class EmptyTracker(Tracker):

    '''A tracker that creates a warning.'''
    slices = ("slice1", "slice2")
    tracks = []

    def __call__(self, track, slice=None):
        return None


class LabeledDataTest(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((("column1", 10),
                          ("column2", 20),))
        elif slice == "slice2":
            return odict((("column1", 20),
                          ("column2", 10),
                          ("column3", 5),))


class StatusTracker(Status):
    tracks = ("track1", "track2", "track3")

    def testTest1(self, track):
        '''test1 passes'''
        return "PASS", 0.5

    def testTest2(self, track):
        '''test2 fails -
        A large test.'''
        return "FAIL", 2

    def testTest3(self, track):
        '''test3 gives a warning'''
        return "WARN", "a string"

    def testTest4(self, track):
        '''test4 is not available/applicable'''
        return "NA", None


class MatrixTracker(Tracker):

    '''returns matrices.'''

    tracks = 'example1', 'example2'

    def __call__(self, track):

        if track == 'example1':
            matrix = numpy.arange(0, 10)
        elif track == 'example2':
            matrix = numpy.arange(10, 0, -1)

        matrix.shape = (2, 5)

        r = {'rows': list(map(str, range(0, 2))),
             'columns': list(map(str, range(0, 5))),
             'matrix': matrix}

        return r


def getSingleValue():
    return 12

IMAGEDIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", "images")


class DataWithImagesExample(Tracker):
    tracks = ("all",)

    def __call__(self, track):

        images = glob.glob(os.path.join(IMAGEDIR, "*.png"))

        data = odict()
        data["numbers"] = range(len(images))
        data["images"] = [".. image:: %s" % x for x in images]

        return data


class VennTracker(Tracker):

    tracks = ('two-circle', 'three-circle')
    slices = ('data1', 'data2')

    def __call__(self, track, slice):

        if slice == 'data1':
            f = 10
        else:
            f = 1

        if track == 'two-circle':
            return {'01': 10, '10': 20 * f, '11': 5 * f,
                    'labels': ("SetA", "SetB")}

        elif track == 'three-circle':
            return {'001': 10, '010': 20 * f, '100': 5 * f,
                    '011': 10, '110': 20 * f, '101': 5 * f,
                    '111': 10,
                    'labels': ("SetA", "SetB", "SetC")}



from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict
from random import randint
import pandas


class ReturnValue(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return randint(0,100)


class ReturnValueWithSlice(Tracker):
    tracks = ("track1", "track2", "track3")
    slices = ("slice1", "slice2")

    def __call__(self, track, slice):
        return randint(0,100)


class ReturnValueInDictionary(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice):

        if slice == "slice1":
            return odict((("column1", randint(0,100)),
                          ("column2", randint(0,100)),))
        elif slice == "slice2":
            return odict((("column1", randint(0,100)),
                          ("column2", randint(0,100)),
                          ("column3", randint(0,100)),))

class ReturnArray(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return [randint(0, 20) for x in range(10)]


class ReturnVariableLengthArray(Tracker):
    tracks = ("track1", "track2", "track3")
    def __call__(self, track):
        lengths = {'track1': 10,
                   'track2': 12,
                   'track3': 14}
        return [randint(0, 20) for x in range(lengths[track])]


class ReturnArrayWithSlice(Tracker):
    tracks = ("track1", "track2", "track3")
    slices = ("slice1", "slice2")

    def __call__(self, track, slice=None):
        return [randint(0, 20) for x in range(10)]


class ReturnVariableLengthArrayWithSlice(Tracker):
    tracks = ("track1", "track2", "track3")
    slices = ("slice1", "slice2")

    def __call__(self, track, slice=None):
        lengths = {'track1': 10,
                   'track2': 12,
                   'track3': 14}
        return [randint(0, 20) for x in range(lengths[track])]


class ReturnDataFrameSimple(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return pandas.DataFrame(
            [randint(0, 20) for x in range(20)])


class ReturnDataFrameWithColumnLabel(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return pandas.DataFrame(
            {'column1': [randint(0, 20) for x in range(20)],
             'column2': [randint(0, 20) for x in range(20)]})


class ReturnDataFrameWithColumnLabels(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return pandas.DataFrame(
            [randint(0, 20) for x in range(20)],
            columns=[track])


class ReturnDataFrameWithIndex(Tracker):
    tracks = ("track1", "track2", "track3")

    def __call__(self, track):
        return pandas.DataFrame(
            [randint(0, 20) for x in range(20)],
            index=['index_%s' % track for x in range(20)])






'''Configuration values in CGATReport.

This module sets some default variables and
neads the sphinx conf.py file.
'''

import os

SEPARATOR = "@"

HTML_IMAGE_FORMAT = ('main', 'png', 80)
LATEX_IMAGE_FORMAT = ()  # ('pdf', 'pdf' 50)

ADDITIONAL_FORMATS = [
    ('hires', 'hires.png', 200),
    # ('eps', 'eps', 50),
    # ('svg', 'svg', 50),
    # ('pdf', 'pdf', 50),
]

# import conf.py
if os.path.exists("conf.py"):
    try:
        exec(compile(open("conf.py").read(), "conf.py", 'exec'))
    except ValueError:
        pass




import os
import sys

from sqlalchemy import *
from sqlalchemy.ext.sqlsoup import SqlSoup

# ignore unknown type BigInt warnings
try:
    db = create_engine(sql_backend)
    db.echo = False
except NameError:
    db = None

if db:
    pass
    #import warnings
    # with warnings.catch_warnings():
    #    warnings.simplefilter("ignore")
    #    metadata = MetaData(db, reflect = True)


def getTables():
    return metadata.sorted_tables


def getTable(name):
    """return table with name *name*."""
    for table in metadata.sorted_tables:
        if table.name == name:
            return table
    raise IndexError("table %s no found" % name)


def execute(stmt):
    return db.execute(stmt)


def getValue(stmt):
    """return a single value from an SQL statement.

    This function will return the first value in the first row
    from an SELECT statement.
    """
    return execute(stmt).fetchone()[0]


def getValues(stmt):
    """return all results from an SQL statement.

    This function will return the first value in each row
    from an SELECT statement.
    """
    return [x[0] for x in execute(stmt).fetchall()]


def getAll(stmt):
    """return all results from an SQL statement.
    """
    return execute(stmt).fetchall()



#!/bin/env python

"""
cgatreport-build
==================

:command:`cgatreport-build` is a pre-processor for restructured
texts. It implements parallel data gathering to speed up the
sphinx document creation process. It is invoked by simply prefixing
the:command:`sphinx` command line::

   cgatreport-build [OPTIONS] sphinx [SPHINX-OPTIONS]

The full list of command line options is listed by suppling:option:`-h/--help`
on the command line.

**-a/--num-jobs** number of jobs
    Number of jobs to start for parallel pre-processing.

**-v/--verbose** verbosity level
    Increase the number of status messages displayed.

"""

import sys
import os
import re
import optparse
import traceback
import hashlib
import subprocess
import logging
import time
import collections
USAGE = """python %s [OPTIONS] args

build a sphinx report.

Building proceeds in three phases.

"""

from CGATReport import report_directive, gallery, clean, Utils

from CGATReport.Component import *

from CGATReport import Logger

try:
    from multiprocessing import Process
    from multiprocessing import Pool, Queue
except ImportError:
    from threading import Thread as Process

# import conf.py for source_suffix
if not os.path.exists("conf.py"):
    raise IOError("could not find conf.py")

exec(compile(open("conf.py").read(), "conf.py", 'exec'))

RST_TEMPLATE = """.. _%(label)s:

.. render:: %(tracker)s
:render: %(renderer)s
   %(options)s

   %(caption)s
"""

# logging.basicConfig(
#     level=logging.DEBUG,
#     format='%(asctime)s %(levelname)s %(message)s',
#     stream = open(Component.LOGFILE, "a") )


class ReportBlock:

    '''quick and dirty parsing of rst of a report block.'''

    def __init__(self):
        self.mLines = []
        self.mOptions = {}
        self.mArguments = None
        self.mCaption = []

    def append(self, v):
        s = v.strip()
        if s.startswith(".. report::"):
            self.mArguments = re.match(".. report::\s+(\S+)", s).groups()
        else:
            s = re.match(":(\S+):\s*(\S*)", v.strip())
            if s:
                key, value = s.groups()
                self.mOptions[key] = value
            else:
                self.mCaption.append(v)
        self.mLines.append(v)


def run(work):
    """run a set of worker jobs.
    """

    try:
        for f, lineno, b, srcdir, builddir in work:
            ff = os.path.abspath(f)
            debug("build.run: profile: started: rst: %s:%i" % (ff, lineno))

            report_directive.run(b.mArguments,
                                 b.mOptions,
                                 lineno=lineno,
                                 content=b.mCaption,
                                 state_machine=None,
                                 document=ff,
                                 srcdir=srcdir,
                                 builddir=builddir)

            debug("build.run: profile: finished: rst: %s:%i" % (ff, lineno))

        return None
    except:
        exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        exception_stack = traceback.format_exc(exceptionTraceback)
        exception_name = exceptionType.__module__ + \
            '.' + exceptionType.__name__
        exception_value = str(exceptionValue)
        return (exception_name, exception_value, exception_stack)


def rst_reader(infile):
    """parse infile and extract the:render: block."""

    result = None
    keep = 0
    lineno = 0
    for line in infile:
        lineno += 1
        if line.startswith(".. report::"):
            if result:
                yield lineno, result
            keep = True
            result = ReportBlock()
            result.append(line)
        else:
            if keep:
                if re.match("^\S", line):
                    keep = False
                else:
                    result.append(line)

    if result:
        yield lineno, result


def getBlocksFromRstFile(rst_file):

    blocks = []
    try:
        infile = open(rst_file, "r")
    except IOError:
        print("could not open %s - skipped" % rst_file)
        return blocks

    for lineno, rst_block in rst_reader(infile):
        blocks.append((lineno, rst_block))
    infile.close()
    return blocks


class timeit:

    """simple timing+logging decorator"""

    def __init__(self, stage):
        self.mStage = stage

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            start = time.time()
            print("CGATReport: phase %s started" % (self.mStage))
            sys.stdout.flush()
            result = func(*args, **kwargs)
            print("CGATReport: phase %s finished in %i seconds" %
                  (self.mStage, time.time() - start))
            sys.stdout.flush()
            return result
        return wrapped


@timeit("getDirectives")
def getDirectives(options, args, sourcedir):
    ''' getting directives.
    '''
    rst_files = []
    for root, dirs, files in os.walk(sourcedir):
        for f in files:
            if f.endswith(source_suffix):
                rst_files.append(os.path.join(root, f))
    return rst_files


@timeit("buildPlots")
def buildPlots(rst_files, options, args, sourcedir):
    '''build all plot elements and tables.

    This can be done in parallel to some extent.
    '''
    info("building plot elements started")

    # build work. Group trackers of the same name together as
    # the python shelve module does not allow concurrent write
    # access and the cache files will get mangled.
    work_per_tracker = collections.defaultdict(list)
    for f in rst_files:
        for lineno, b in getBlocksFromRstFile(f):
            work_per_tracker[b.mArguments].append((f,
                                                   lineno,
                                                   b,
                                                   sourcedir,
                                                   "."))

    work = []
    for tracker, vals in work_per_tracker.items():
        work.append(vals)

    if len(work) == 0:
        return

    if options.num_jobs > 1:
        logQueue = Queue(100)
        handler = Logger.MultiProcessingLogHandler(
            logging.FileHandler(os.path.abspath(LOGFILE), "w"), logQueue)
    else:
        handler = logging.FileHandler(os.path.abspath(LOGFILE), "w")

    handler.setFormatter(
        logging.Formatter(LOGGING_FORMAT))

    logging.getLogger('').addHandler(handler)
    logging.getLogger('').setLevel(options.loglevel)

    info('starting %i jobs on %i work items' % (options.num_jobs, len(work)))
    debug("build.py: profile: started: 0 seconds")

    if options.num_jobs > 1:
        pool = Pool(options.num_jobs)
        # todo: async execution with timeouts
        # res = pool.map_async(run, work)
        errors = pool.map(run, work)
        pool.close()
        pool.join()
    else:
        errors = []
        for w in work:
            errors.append(run(w))

    errors = [e for e in errors if e]

    if errors:
        print("CGATReport caught %i exceptions" % (len(errors)))
        print("## start of exceptions")
        for exception_name, exception_value, exception_stack in errors:
            print(exception_stack)
        print("## end of exceptions")
        sys.exit(1)

    if options.num_jobs > 1:
        counts = handler.getCounts()

        print(("CGATReport: messages: %i critical, %i errors, "
               "%i warnings, %i info, %i debug") %
              (counts["CRITICAL"],
               counts["ERROR"],
               counts["WARNING"],
               counts["INFO"],
               counts["DEBUG"]))

    logging.shutdown()


@timeit("cleanTrackers")
def cleanTrackers(rst_files, options, args):
    '''instantiate trackers and get code.'''
    trackers = set()
    for f in rst_files:
        for lineno, b in getBlocksFromRstFile(f):
            trackers.add(b.mArguments[0])

    ntested, ncleaned, nskipped = 0, 0, 0
    for reference in trackers:

        try:
            code, tracker, tracker_path = report_directive.makeTracker(reference)
        except AttributeError:
            # ignore missing trackers
            nskipped += 1
            continue

        new_codehash = hashlib.md5("".join(code)).hexdigest()
        (basedir, fname, basename, ext,
         outdir, codename, notebookname) = report_directive.buildPaths(
             reference)
        codefilename = os.path.join(outdir, codename)
        ntested += 1
        if not os.path.exists(codefilename):
            nskipped += 1
            continue
        old_codehash = hashlib.md5(
            "".join(open(codefilename, "r").readlines())).hexdigest()
        if new_codehash != old_codehash:
            removed = clean.removeTracker(reference)
            removed.extend(clean.removeText(reference))
            print("code has changed for %s: %i files removed" %
                  (reference, len(removed)))
            ncleaned += 1
    print("CGATReport: %i Trackers changed (%i tested, %i skipped)" %
          (ncleaned, ntested, nskipped))


def runCommand(command):
    try:
        retcode = subprocess.call(command, shell=True)
        if retcode < 0:
            warn("child was terminated by signal %i" % -retcode)
    except OSError(msg):
        fail("execution of %s failed: %s" % (cmd, msg))
        raise


@timeit("buildGallery")
def buildGallery(options, args):
    """construct the gallery page.
    """
    gallery.main()


@timeit("buildDocument")
def buildDocument(options, args):
    """construct the documents. This is simply done
    by calling sphinx-build.
    """
    runCommand("%s" % " ".join(args))


@timeit("buildLog")
def buildLog(options, args):
    """construct pages with the error log, stats
    and the build history.
    """
    runCommand("%s" % " ".join(args))


def main(argv=None):

    if argv is None:
        argv = sys.argv

    print("CGATReport: version %s started" % str("$Id$"))
    t = time.time()

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-a", "--num-jobs", dest="num_jobs", type="int",
                      help="number of parallel jobs to run [default=%default]")

    parser.add_option("-v", "--verbose", dest="loglevel", type="int",
                      help="loglevel. The higher, the more output "
                      "[default=%default]")

    parser.set_defaults(num_jobs=2,
                        loglevel=10,)

    parser.disable_interspersed_args()

    (options, args) = parser.parse_args()

    assert args[0].endswith(
        "sphinx-build"), "command line should contain sphinx-build"

    sphinx_parser = optparse.OptionParser(
        version="%prog version: $Id$", usage=USAGE)
    sphinx_parser.add_option("-b", type="string")
    sphinx_parser.add_option("-a")
    sphinx_parser.add_option("-E")
    sphinx_parser.add_option("-t", type="string")
    sphinx_parser.add_option("-d", type="string")
    sphinx_parser.add_option("-c", type="string")
    sphinx_parser.add_option("-C")
    sphinx_parser.add_option("-D", type="string")
    sphinx_parser.add_option("-A", type="string")
    sphinx_parser.add_option("-n")
    sphinx_parser.add_option("-Q")
    sphinx_parser.add_option("-q")
    sphinx_parser.add_option("-w", type="string")
    sphinx_parser.add_option("-W")
    sphinx_parser.add_option("-P")

    (sphinx_options, sphinx_args) = sphinx_parser.parse_args(args[1:])

    sourcedir = sphinx_args[0]

    rst_files = getDirectives(options, args, sourcedir)

    cleanTrackers(rst_files, options, args)

    buildPlots(rst_files, options, args, sourcedir)

    # buildGallery(options, args)

    # buildLog(options, args)

    buildDocument(options, args)

    print("CGATReport: finished in %i seconds" % (time.time() - t))

    debug("build.py: profile: finished: %i seconds" % (time.time() - t))

if __name__ == "__main__":
    sys.exit(main())



from CGATReport.Tracker import TrackerSQL


class ExpressionLevel(TrackerSQL):

    """Expression level measurements."""
    pattern = "(.*)_data$"

    def __init__(self, *args, **kwargs):
        TrackerSQL.__init__(self,
                            *args,
                            backend="sqlite:///./csvdb",
                            **kwargs)

    def __call__(self, track):
        statement = "SELECT expression FROM %s_data" % track
        data = self.getValues(statement)
        return {"expression": data}


class ExpressionLevelWithSlices(ExpressionLevel):

    """Expression level measurements."""

    slices = ("housekeeping", "regulation")

    def __call__(self, track, slice):
        if not slice:
            where = ""
        else:
            where = "WHERE function = '%s'" % slice
        statement = "SELECT expression FROM %s_data %s" % (track, where)
        data = self.getValues(statement)
        return {"expression": data}



import re
import os
from docutils import nodes, utils
from docutils.parsers.rst import roles

from CGATReport import Utils

default_settings = {
    'pubmed_url': "http://www.ncbi.nlm.nih.gov/pubmed/%i",
    'biogps_url': "http://biogps.org/?query=%s",
    'ucsc_url':
    "http://genome.ucsc.edu/cgi-bin/hgTracks?db=%s&position=%s",
    'ensembl_generic_search':
    'http://www.ensembl.org/Multi/Search/Results?q=%s',
    'ensembl_species_search':
    'http://www.ensembl.org/%s/Search/Results?q=%s'}


def writeCode(class_name, code, inliner):
    '''write code of class to file.

    returns URI of written code.
    '''
    document = inliner.document.current_source

    reference = class_name

    # root of document tree
    srcdir = setup.srcdir

    # get the directory of the rst file
    rstdir, rstfile = os.path.split(document)

    (basedir, fname, basename, ext, outdir, codename,
     notebookname) = Utils.buildPaths(reference)

    # path to root relative to rst
    rst2srcdir = os.path.join(os.path.relpath(srcdir, start=rstdir), outdir)

    # output code
    linked_codename = re.sub("\\\\", "/", os.path.join(rst2srcdir, codename))
    if code and basedir != outdir:
        outfile = open(os.path.join(outdir, codename), "w")
        for line in code:
            outfile.write(line)
        outfile.close()

    return linked_codename


def pubmed_role(role, rawtext, text, lineno, inliner,
                options={}, content=[]):
    '''insert a link to pubmed into the text.'''
    try:
        pmid = int(text)
        if pmid <= 0:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'pmid number must be a number greater than or equal to 1; '
            '"%s" is invalid.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    # Base URL mainly used by inliner.rfc_reference, so this is correct:
    ref = default_settings["pubmed_url"] % pmid
    # in example, but deprecated
    # set_classes(options)
    node = nodes.reference(rawtext, 'PMID ' + utils.unescape(text), refuri=ref,
                           **options)

    return [node], []


def biogps_role(role, rawtext, text, lineno, inliner,
                options={}, content=[]):
    '''insert a link to biogps into the text.'''
    ref = default_settings["biogps_url"] % text
    node = nodes.reference(rawtext,
                           utils.unescape(text),
                           refuri=ref,
                           **options)

    return [node], []


def ucsc_role(role, rawtext, text, lineno, inliner,
              options={}, content=[]):
    '''insert a link to biogps into the text.'''
    if '@' not in text:
        msg = inliner.reporter.error(
            'the ucsc role expects a database and a coordinate '
            'separated by "@" '
            '"%s" is invalid.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    database, coordinates = text.split('@', 1)

    ref = default_settings["ucsc_url"] % (database, coordinates)
    node = nodes.reference(rawtext,
                           utils.unescape(text),
                           refuri=ref,
                           **options)

    return [node], []


def ensembl_role(role, rawtext, text, lineno, inliner,
                 options={}, content=[]):
    '''insert a link to biogps into the text.'''
    if '@' in text:
        database, query = text.split('@', 1)
        ref = default_settings["ensembl_species_search"] % (
            database, query)
    else:
        ref = default_settings["ensembl_generic_search"] % text
    node = nodes.reference(rawtext,
                           utils.unescape(text),
                           refuri=ref,
                           **options)

    return [node], []


def param_role(role, rawtext, text, lineno, inliner,
               options={}, content=[]):
    '''inserts a member variable of a tracker class in the text.'''

    parts = text.split(".")

    if len(parts) < 2:
        msg = inliner.reporter.error(
            ':param: should be in class.value format '
            ': "%s" is invalid.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    class_name, parameter_name = ".".join(parts[:-1]), parts[-1]

    try:
        code, tracker, tracker_path = Utils.makeTracker(class_name)
    except AttributeError:
        tracker = None

    if not tracker:
        msg = inliner.reporter.error(
            ':param: can not find class '
            '"%s".' % class_name, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    try:
        value = getattr(tracker, parameter_name)
    except AttributeError:
        msg = inliner.reporter.error(
            ':param: can not find variable %s in '
            ': "%s" is invalid -tracker=%s.' %
            (parameter_name, class_name, str(tracker)), line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    linked_codename = writeCode(class_name, code, inliner)

    node = nodes.reference(rawtext,
                           utils.unescape(str(value)),
                           refuri=linked_codename,
                           **options)

    return [node], []


def value_role(role, rawtext, text, lineno, inliner,
               options={}, content=[]):
    '''insert a single value from a tracker into text.'''

    class_name = text

    try:
        code, tracker, tracker_path = Utils.makeTracker(class_name)
    except (AttributeError, ImportError):
        tracker = None

    if not tracker:
        msg = inliner.reporter.error(
            ':value: can not find class '
            '"%s".' % class_name, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    # Python 2/3
    try:
        value = str(tracker())
    except TypeError as msg:
        print("python 3 problem: %s: tracker=%s" % (msg, str(tracker())))

    linked_codename = writeCode(class_name, code, inliner)

    # Base URL mainly used by inliner.rfc_reference, so this is correct:
    # ref = inliner.document.settings.rfc_base_url + inliner.rfc_url % value
    # in example, but deprecated
    # set_classes(options)
    # node = nodes.literal(rawtext,
    #                    utils.unescape(str(value)),
    #                   **options)
    node = nodes.reference(rawtext,
                           utils.unescape(str(value)),
                           refuri=linked_codename,
                           **options)

    return [node], []


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir
    setup.srcdir = app.srcdir

roles.register_local_role('pmid', pubmed_role)
roles.register_local_role('biogps', biogps_role)
roles.register_local_role('ucsc', ucsc_role)
roles.register_local_role('ensembl', ensembl_role)
roles.register_local_role('param', param_role)
roles.register_local_role('value', value_role)



'''Script to fill csvdb database with example data.'''

import random, os
from subprocess import Popen

cmd = 'sqlite3 csvdb "%s" '

def e( stmt ):
    p = Popen( cmd % stmt, shell=True )
    sts = os.waitpid(p.pid, 0)

e("CREATE TABLE experiment1_data (gene_id TEXT, function TEXT, expression INT)")

for x in range(0,200,2):
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" % ("experiment1_data", x, "housekeeping", random.gauss( 41, 5)) )
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" % ("experiment1_data", x+1, "regulation", random.gauss( 11, 5)) )

e("CREATE TABLE experiment2_data (gene_id TEXT, function TEXT, expression INT)")

for x in range(0,200,2):
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" % ("experiment2_data", x, "housekeeping", random.gauss( 40, 5)) )
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" % ("experiment2_data", x+1, "regulation", random.gauss( 10, 5)) )





'''basic module for CGATReport actors.
'''

import os
import sys
import pkg_resources
from logging import warn, debug, info, critical, error
import logging
import collections
from docutils.parsers.rst import directives

LOGFILE = "cgatreport.log"
LOGGING_FORMAT = '%(asctime)s %(levelname)s %(message)s'

logging.basicConfig(
    level=logging.DEBUG,
    format=LOGGING_FORMAT,
    stream=open(LOGFILE, "a"))


class Component(object):

    '''base class for CGATReport components.
    '''

    # options exported to sphinx
    options = ()

    def __init__(self, *args, **kwargs):
        pass

    def debug(self, msg):
        debug("disp%s: %s" % (id(self), msg))

    def warn(self, msg):
        warn("disp%s: %s" % (id(self), msg))

    def info(self, msg):
        logging.info("disp%s: %s" % (id(self), msg))

    def error(self, msg):
        error("disp%s: %s" % (id(self), msg))

    def critical(self, msg):
        critical("disp%s: %s" % (id(self), msg))

# plugins are only initialized once they are called
# for in order to remove problems with cyclic imports
plugins = None
options = None
ENTRYPOINT = 'CGATReport.plugins'


def init_plugins():

    info("initialising plugins")
    try:
        pkg_resources.working_set.add_entry(cgatreport_plugins)
        pkg_env = pkg_resources.Environment(cgatreport_plugins)
    except NameError:
        pkg_env = pkg_resources.Environment()

    plugins = collections.defaultdict(dict)
    for name in pkg_env:
        egg = pkg_env[name][0]
        egg.activate()
        modules = []
        for name in egg.get_entry_map(ENTRYPOINT):
            entry_point = egg.get_entry_info(ENTRYPOINT, name)
            cls = entry_point.load()
            if not hasattr(cls, 'capabilities'):
                cls.capabilities = []

            for c in cls.capabilities:
                plugins[c][name] = cls

    if len(plugins) == 0:
        warn("did not find any plugins")
    else:
        debug("found plugins: %i capabilites and %i plugins" %
              (len(plugins), sum([len(x) for x in list(plugins.values())])))

    return plugins


def getPlugins(capability=None):
    global plugins
    if not plugins:
        plugins = init_plugins()
    if capability is None:
        return plugins
        result = set()
        for p in plugins.values():
            for plugin in p:
                result.add(plugin)
        return list(result)
    else:
        return plugins.get(capability, {})


def getOptionMap():
    global options

    if options is None:
        options = {}
        for section, plugins in getPlugins().items():
            options[section] = {}
            for name, cls in plugins.items():
                try:
                    options[section].update(dict(cls.options))
                except AttributeError:
                    pass
        options["dispatch"] = {
            'groupby': directives.unchanged,
            'tracks': directives.unchanged,
            'slices': directives.unchanged,
            'layout': directives.unchanged,
            'restrict': directives.unchanged,
            'exclude': directives.unchanged,
            'nocache': directives.flag,
        }

        # options used in trackers
        options["tracker"] = {
            'regex': directives.unchanged,
            'glob': directives.unchanged,
        }

        options["display"] = {
            # general image options
            'alt': directives.unchanged,
            'height': directives.length_or_unitless,
            'width': directives.length_or_percentage_or_unitless,
            'scale': directives.nonnegative_int,
            'class': directives.class_option,
            # 'align': align,
            # options for cgatreport
            'render': directives.unchanged,
            'transform': directives.unchanged,
            'display': directives.unchanged,
            'extra-formats': directives.unchanged,
        }

    return options


def getOptionSpec():
    '''build option spec for sphinx

    This method returns a flattened:var:`options`.
    '''
    o = getOptionMap()
    r = {}
    for x, xx in o.items():
        r.update(xx)

    # add the primary actor options
    r["render"] = directives.unchanged
    r["transform"] = directives.unchanged

    return r



#!/bin/env python

"""
cgatreport-serve
==================

:command:`cgatreport-serve` starts a minimalist web server that permits
the user to interact with some of the elements in a sphinxport document. In particular,
it enables the ``data`` element permitting the download of raw data.

To start the server, type::

   cgatreport-serve

in a cgatreport directory. Once a server has started, the documents can
be accessed at from the same host as:

   http://localhost:8080/static

.. note::
   This feature has not been thoroughly tested. In particular, the security implications
   are unclear and the tools should best be used behind a firewall only.

The full list of command line options is listed by suppling:option:`-h/--help`
on the command line.

The options are:

**-t/--html** html dir
   Directory of the report in html format. The default is "_build/html".
   The html directory should contain a file:file:`index.html`. A link
   called ``static`` will be created in the current directory.

**-p/--port** port
   The port to use. The default port is ``8080``. If the port is changed,
   the URL in the example above needs to be changed accordingly.

**-a/--action** action
   Action to perform. The default is to ``start`` the server. Other
   actions are ``stop`` to stop and ``restart`` to restart the server.
"""

import sys
import os
import imp
import io
import re
import types
import glob
import optparse
import shutil

USAGE = """python %s [OPTIONS]

start a cgatreport server

""" % sys.argv[0]


import web

from CGATReport import Utils
from CGATReport import Cache
from CGATReport import DataTree
from collections import OrderedDict as odict


urls = ('/data/(.*)', 'DataTable',
        '/index/(.*)', 'Index')

# expose zip within templates
global_vars = {'zip': zip}

render = web.template.render(
    '%s/' % Utils.getTemplatesDir(), globals=global_vars)

app = web.application(urls, globals())


class DataTable:

    '''render data retrieved from cache as a table.'''

    def GET(self, tracker):

        cache = Cache.Cache(tracker, mode="r")
        data = DataTree.fromCache(cache)
        table, row_headers, col_headers = DataTree.tree2table(data)

        return render.data_table(table, row_headers, col_headers)


def main():

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-p", "--port", dest="port", type="int",
                      help="the port to use [default=%default]")

    parser.add_option("-t", "--html", dest="htmldir", type="string",
                      help="html directory [default=%default]")

    parser.add_option("-a", "--action", dest="action", type="choice",
                      choices=("start", "stop", "restart"),
                      help="action to perform [default=%default]")

    parser.set_defaults(
        htmldir="_build/html",
        port=8080,
        action="start",
    )

    (options, args) = parser.parse_args()

    if options.action == "start":
        # create static link
        if not os.path.exists(options.htmldir):
            raise IOError(
                "html directory %s does not exist, use --html option" % options.htmldir)
        if not os.path.exists(os.path.join(options.htmldir, "index.html")):
            raise IOError(
                "html directory %s has no 'index.html'" % options.htmldir)

        if not os.path.exists("static"):
            os.symlink(options.htmldir, "static")
        else:
            if not os.path.exists(os.path.join("static", "index.html")):
                raise IOError("directory %s has no 'index.html'" % "static")

        # set port
        sys.argv = [sys.argv[0], str(options.port)]
        app.run()
    elif options.action == "stop":
        raise NotImplementedError("stop not implemented - try ctrl-c")
    elif options.action == "restart":
        raise NotImplementedError("restart not implemented - try ctrl-c")

if __name__ == "__main__":
    sys.exit(main())
